##
# File:  DataExchange.py
# Date:  19-Sep-2012
#
# Updates:
#
#  28-Feb-2013 jdw incorporate PathInfo() class
#  05-Mar-2013 jdw fetch() should not strip version information for output files.
#  08-Mar-2013 jdw add method to separately set path for 'session' storage.
#  29-Jun-2014 jdw refactor add getVersionFileList() and getPartitionFileList()
#   5-Jul-2014 jdw  getContentTypeFileList() and getMiscFileList()
#   5-Jan-2015 jdw add siteId as an optional argument to the constructor -
#  16-Feb-2017 jdw add setFileSource() method ...
##
"""
 Implements common data exchange operations including: moving annotation data files between session
and workflow storage, accessing files in workflow directories,  and routine file maintenance operations.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.09"

import sys
import os
import shutil
import traceback
import glob
from datetime import datetime

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.io.locator.PathInfo import PathInfo


class DataExchange(object):

    """
     Implements common data exchange operations
     including: moving annotation data files between session
     and workflow storage, accessing files in workflow directories,
     and routine file maintenance operations.

    """

    def __init__(self, reqObj=None, depDataSetId=None, wfInstanceId=None, fileSource="archive", siteId=None, verbose=False, log=sys.stderr):

        self.__reqObj = reqObj
        self.__depDataSetId = depDataSetId
        self.__wfInstanceId = wfInstanceId
        self.__fileSource = fileSource
        self.__verbose = verbose
        self.__lfh = log
        #
        self.__debug = False
        self.__inputSessionPath = None
        #
        self.__setup(siteId=siteId)

    def __setup(self, siteId=None):
        if siteId is not None:
            self.__siteId = siteId
        else:
            self.__siteId = self.__reqObj.getValue("WWPDB_SITE_ID")

        self.__sessionObj = self.__reqObj.getSessionObj()
        self.__sessionPath = self.__sessionObj.getPath()

        self.__cI = ConfigInfo(self.__siteId)
        self.__pI = PathInfo(siteId=self.__siteId, sessionPath=self.__sessionPath, verbose=self.__verbose, log=self.__lfh)

        #
        if self.__debug:
            self.__lfh.write("+DataExchange.__setup() - session id   %s\n" % (self.__sessionObj.getId()))
            self.__lfh.write("+DataExchange.__setup() - session path %s\n" % (self.__sessionObj.getPath()))

            self.__lfh.write("+DataExchange.__setup() - data set %s  instance %s file source %s\n" % (self.__depDataSetId, self.__wfInstanceId, self.__fileSource))
            self.__pI.setDebugFlag(flag=self.__debug)

    def setFileSource(self, fileSource):
        """  Override fileSource="archive"
        """
        self.__fileSource = fileSource

    def setInputSessionPath(self, inputSessionPath=None):
        """  Override the path to files with fileSource="session"
        """
        self.__inputSessionPath = inputSessionPath

    def purgeLogs(self):
        archivePath = self.__cI.get("SITE_ARCHIVE_STORAGE_PATH")
        dirPath = os.path.join(archivePath, "archive", self.__depDataSetId, "log")
        if self.__verbose:
            self.__lfh.write("+DataExchange.purgeLogs() - purging logs in directory  %s\n" % (dirPath))

        if os.access(dirPath, os.W_OK):
            fpattern = os.path.join(dirPath, "*log")
            if self.__verbose:
                self.__lfh.write("+DataExchange.purgeLogs() - purging pattern is %s\n" % (fpattern))

            pthList = glob.glob(fpattern)
            if self.__verbose:
                self.__lfh.write("+DataExchange.purgeLogs() candidate path length is %d\n" % len(pthList))
            #
            for pth in pthList:
                try:
                    os.remove(pth)
                except:  # noqa: E722 pylint: disable=bare-except
                    pass
            #
        return pthList

    def reversePurge(self, contentType, formatType="pdbx", partitionNumber=1):
        fn = self.__getArchiveFileName(contentType=contentType, formatType=formatType, version="none", partitionNumber=partitionNumber)

        archivePath = self.__cI.get("SITE_ARCHIVE_STORAGE_PATH")
        dirPath = os.path.join(archivePath, "archive", self.__depDataSetId)
        if self.__verbose:
            self.__lfh.write("+DataExchange.__setup() - purging in directory  %s\n" % (dirPath))

        if len(dirPath) < 2:
            return []
        fpattern = os.path.join(dirPath, fn + ".V*")
        if self.__verbose:
            self.__lfh.write("+DataExchange.__setup() - purging pattern is %s\n" % (fpattern))

        pthList = glob.glob(fpattern)
        if self.__verbose:
            self.__lfh.write("+DataExchange.__reversePurge() candidate length is %d\n" % len(pthList))
        #
        fList = []
        for pth in pthList:
            if not pth.endswith(".V1"):
                fList.append(pth)

        for pth in fList:
            try:
                os.remove(pth)
            except:  # noqa: E722 pylint: disable=bare-except
                pass
            #
        return fList

    def removeWorkflowDir(self):
        if (self.__depDataSetId is not None) and self.__depDataSetId.startswith("D_") and (len(self.__depDataSetId) > 7):
            workflowPath = self.__cI.get("SITE_WORKFLOW_STORAGE_PATH")
            dirPath = os.path.join(workflowPath, "workflow", self.__depDataSetId)
            if os.access(dirPath, os.W_OK):
                shutil.rmtree(dirPath)
                return True
            else:
                return False
        else:
            return False

    def createArchiveDir(self, purgeFlag=True):
        """ Create new the archive directory if this is needed.

        """

        if self.__verbose:
            self.__lfh.write("+DataExchange.export() creating archive directory for data set %s\n" % self.__depDataSetId)

        try:
            archivePath = self.__cI.get("SITE_ARCHIVE_STORAGE_PATH")
            dirPath = os.path.join(archivePath, "archive", self.__depDataSetId)

            if not os.access(dirPath, os.W_OK):
                if self.__verbose:
                    self.__lfh.write("+DataExchange.createArchiveDir() creating archive directory path %s\n" % dirPath)
                os.makedirs(dirPath)
                return True
            else:
                if purgeFlag:
                    if self.__verbose:
                        self.__lfh.write("+DataExchange.export() existing archive directory path purged: %s\n" % dirPath)
                    shutil.rmtree(dirPath)
                    os.makedirs(dirPath)
                    return True
                else:
                    if self.__verbose:
                        self.__lfh.write("+DataExchange.export() archive directory exists: %s\n" % dirPath)
                    return False
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
            return False

    def fetch(self, contentType, formatType, version="latest", partitionNumber=1):
        """ Copy the input content object into the current session directory (session naming semantics follow source file object)

            Return the full path of the copied file or None

        """
        inpFilePath = self.__getFilePath(fileSource=self.__fileSource, contentType=contentType, formatType=formatType, version=version, partitionNumber=partitionNumber)
        if self.__verbose:
            self.__lfh.write("+DataExchange.fetch() source type %s format %s version %s path %s\n" % (contentType, formatType, version, inpFilePath))

        try:
            if os.access(inpFilePath, os.R_OK):
                (_dirPath, fileName) = os.path.split(inpFilePath)
                # trim of the trailing version -
                # lastIdx=tfileName.rfind(".V")
                # if lastIdx > 0:
                #    fileName=tfileName[:lastIdx]
                # else:
                #    fileName=tfileName
                outFilePath = os.path.join(self.__sessionPath, fileName)
                if self.__verbose:
                    self.__lfh.write("+DataExchange.fetch() destination file path %s\n" % outFilePath)
                shutil.copyfile(inpFilePath, outFilePath)
                return outFilePath
            else:
                if self.__verbose:
                    self.__lfh.write("+DataExchange.fetch() missing input file at path %s\n" % inpFilePath)
                return None
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
            return None

    def export(self, inpFilePath, contentType, formatType, version="latest", partitionNumber=1):
        """ Copy input file to workflow instance or archival storage.

            Return True on success or False otherwise.

        """
        outFilePath = self.__getFilePath(fileSource=self.__fileSource, contentType=contentType, formatType=formatType, version=version, partitionNumber=partitionNumber)
        if self.__verbose:
            self.__lfh.write("+DataExchange.export() destination type %s format %s version %s path %s\n" % (contentType, formatType, version, outFilePath))

        try:
            if os.access(inpFilePath, os.R_OK) and (os.path.getsize(inpFilePath) > 0):
                if self.__verbose:
                    self.__lfh.write("+DataExchange.export() destination file path %s\n" % outFilePath)
                if inpFilePath.endswith(".gz"):
                    self.__copyGzip(inpFilePath, outFilePath)
                else:
                    shutil.copyfile(inpFilePath, outFilePath)
                return True
            else:
                if self.__verbose:
                    self.__lfh.write("+DataExchange.export() missing or zero length input file at path %s\n" % inpFilePath)
                return False
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
            return False

    def __copyGzip(self, inpFilePath, outFilePath):
        """
        """
        try:
            cmd = " gzip -cd  %s > %s " % (inpFilePath, outFilePath)
            os.system(cmd)
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
            return False

    def copyDirToSession(self, dirName):
        """  Replicate the input diretory in the session directory -
        """
        try:
            if self.__fileSource in ["archive", "wf-archive"]:
                pth = self.__pI.getArchivePath(self.__depDataSetId)
            elif self.__fileSource in ["deposit"]:
                pth = self.__pI.getDepositPath(self.__depDataSetId)
            elif self.__fileSource in ["wf-instance"]:
                pth = self.__pI.getInstancePath(self.__depDataSetId, self.__wfInstanceId)
            else:
                return False

            srcPath = os.path.join(pth, dirName)
            if not os.access(srcPath, os.R_OK):
                return False

            dstPath = os.path.join(self.__sessionPath, dirName)
            if not os.path.isdir(dstPath):
                os.makedirs(dstPath, 0o755)
            #
            fPattern = os.path.join(srcPath, "*")
            fpL = filter(os.path.isfile, glob.glob(fPattern))
            for fp in fpL:
                _dN, fN = os.path.split(fp)
                oP = os.path.join(dstPath, fN)
                shutil.copyfile(fp, oP)

            if self.__verbose:
                self.__lfh.write("+DataExchange.copyDirToSession() successful session copy of dirName %s\n" % (dirName))
            return True
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                self.__lfh.write("+DataExchange.copyDirToSession() fails for dirName %s\n" % (dirName))
                traceback.print_exc(file=self.__lfh)
            return False

        return True

    def copyToSession(self, contentType, formatType, version="latest", partitionNumber=1):
        """ Copy the input content object into the session directory using archive naming conventions less version details.

            Return the full path of the session file or None

        """
        inpFilePath = self.__getFilePath(fileSource=self.__fileSource, contentType=contentType, formatType=formatType, version=version, partitionNumber=partitionNumber)
        if self.__debug:
            self.__lfh.write("+DataExchange.copyToSession() source file type %s format %s version %s path %s\n" % (contentType, formatType, version, inpFilePath))

        try:
            outFilePath = None
            if os.access(inpFilePath, os.R_OK):
                fn = self.__getArchiveFileName(contentType, formatType, version="none", partitionNumber=partitionNumber)
                outFilePath = os.path.join(self.__sessionPath, fn)
                if self.__verbose:
                    self.__lfh.write("+DataExchange.copyToSession() content type %s format %s copied to session path %s\n" % (contentType, formatType, outFilePath))
                shutil.copyfile(inpFilePath, outFilePath)
                return outFilePath
            else:
                if self.__debug:
                    self.__lfh.write("+DataExchange.copyToSession() missing input file at path %s\n" % inpFilePath)
                return None
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                if self.__verbose:
                    self.__lfh.write("+DataExchange.copyToSession() Failing for content type %s format %s with session path %s\n" % (contentType, formatType, outFilePath))
                traceback.print_exc(file=self.__lfh)
            return None

    def updateArchiveFromSession(self, contentType, formatType, version="next", partitionNumber=1):
        """ Copy the input content object from the session directory stored using  archive naming conventions less version details
            to archive storage.

            Return the full path of the archive file or None

        """
        fn = self.__getArchiveFileName(contentType, formatType, version="none", partitionNumber=partitionNumber)
        inpFilePath = os.path.join(self.__sessionPath, fn)
        if self.__verbose:
            self.__lfh.write("+DataExchange.updateArchiveDromSession() source file type %s format %s path %s\n" % (contentType, formatType, inpFilePath))

        try:
            if os.access(inpFilePath, os.R_OK):
                outFilePath = self.__getFilePath(fileSource="archive", contentType=contentType, formatType=formatType, version=version, partitionNumber=partitionNumber)
                if self.__verbose:
                    self.__lfh.write("+DataExchange.updateArchiveFromSession() archive destination file path %s\n" % outFilePath)
                shutil.copyfile(inpFilePath, outFilePath)
                return outFilePath
            else:
                if self.__verbose:
                    self.__lfh.write("+DataExchange.updateArchiveFrom() missing session input file at path %s\n" % inpFilePath)
                return None
        except:  # noqa: E722 pylint: disable=bare-except
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
            return None

    ##
    def getVersionFileList(self, fileSource="archive", contentType="model", formatType="pdbx", partitionNumber="1", mileStone=None):
        """
        For the input content object return a list of file versions sorted by modification time.

        Return:
              List of [(file path, modification date string,size),...]

        """
        try:
            if fileSource == "session" and self.__inputSessionPath is not None:
                self.__pI.setSessionPath(self.__inputSessionPath)

            fPattern = self.__pI.getFilePathVersionTemplate(
                dataSetId=self.__depDataSetId,
                wfInstanceId=self.__wfInstanceId,
                contentType=contentType,
                formatType=formatType,
                fileSource=fileSource,
                partNumber=partitionNumber,
                mileStone=mileStone,
            )
            return self.__getFileList([fPattern], sortFlag=True)
        except Exception as e:
            if self.__verbose:
                self.__lfh.write(
                    "+DataExchange.getVersionFileList() failing for data set %s instance %s file source %s error %r\n"
                    % (self.__depDataSetId, self.__wfInstanceId, self.__fileSource, str(e))
                )
                traceback.print_exc(file=self.__lfh)
            return []

    def getPartitionFileList(self, fileSource="archive", contentType="model", formatType="pdbx", mileStone=None):
        """
        For the input content object return a list of file partitions sorted by modification time.

        Return:
              List of [(file path, modification date string,size),...]

        """
        try:
            if fileSource == "session" and self.__inputSessionPath is not None:
                self.__pI.setSessionPath(self.__inputSessionPath)

            fPattern = self.__pI.getFilePathPartitionTemplate(
                dataSetId=self.__depDataSetId, wfInstanceId=self.__wfInstanceId, contentType=contentType, formatType=formatType, fileSource=fileSource, mileStone=mileStone
            )
            tL = self.__getFileList([fPattern], sortFlag=True)
            if self.__debug:
                self.__lfh.write("+DataExchange.getPartionFileList() pattern %r\n" % fPattern)
                self.__lfh.write("+DataExchange.getPartionFileList() file list %r\n" % tL)
            #
            return tL
        except Exception as e:
            if self.__verbose:
                self.__lfh.write(
                    "+DataExchange.getVersionFileList() failing for data set %s instance %s file source %s error %r\n"
                    % (self.__depDataSetId, self.__wfInstanceId, self.__fileSource, str(e))
                )
                traceback.print_exc(file=self.__lfh)
            return []

    def getContentTypeFileList(self, fileSource="archive", contentTypeList=None):
        """
        For the input content object return a list of file versions sorted by modification time.

        Return:
              List of [(file path, modification date string,size),...]

        """
        if contentTypeList is None:
            contentTypeList = ["model"]
        try:
            if fileSource == "session" and self.__inputSessionPath is not None:
                self.__pI.setSessionPath(self.__inputSessionPath)
            fPatternList = []
            for contentType in contentTypeList:
                fPattern = self.__pI.getFilePathContentTypeTemplate(dataSetId=self.__depDataSetId, wfInstanceId=self.__wfInstanceId, contentType=contentType, fileSource=fileSource)

                fPatternList.append(fPattern)
            if self.__debug:
                self.__lfh.write("+DataExchange.getContentTypeFileList() patterns %r\n" % fPatternList)
            return self.__getFileList(fPatternList, sortFlag=True)
        except Exception as e:
            if self.__verbose:
                self.__lfh.write(
                    "+DataExchange.getVersionFileList() failing for data set %s instance %s file source %s error %r\n"
                    % (self.__depDataSetId, self.__wfInstanceId, self.__fileSource, str(e))
                )
                traceback.print_exc(file=self.__lfh)
            return []

    def getMiscFileList(self, fPatternList=None, sortFlag=True):
        if fPatternList is None:
            fPatternList = ["*"]
        return self.__getFileList(fPatternList=fPatternList, sortFlag=sortFlag)

    def getLogFileList(self, entryId, fileSource="archive"):
        if fileSource in ["archive", "wf-archive"]:
            pth = self.__pI.getArchivePath(entryId)
            fpat1 = os.path.join(pth, "*log")
            fpat2 = os.path.join(pth, "log", "*")
            patList = [fpat1, fpat2]
        elif fileSource in ["deposit"]:
            pth = self.__pI.getDepositPath(entryId)
            fpat1 = os.path.join(pth, "*log")
            fpat2 = os.path.join(pth, "log", "*")
            patList = [fpat1, fpat2]
        else:
            return []
        return self.__getFileList(fPatternList=patList, sortFlag=True)

    def __getFileList(self, fPatternList=None, sortFlag=True):
        """
        For the input glob compatible file pattern produce a file list sorted by modification date.

        If sortFlag is set then file list is sorted by modification date (e.g. recently changes first)

        Return:
              List of [(file path, modification date string, KBytes),...]

        """
        if fPatternList is None:
            fPatternList = ["*"]
        rTup = []
        try:
            files = []
            for fPattern in fPatternList:
                if fPattern is not None and len(fPattern) > 0:
                    files.extend(filter(os.path.isfile, glob.glob(fPattern)))

            file_date_tuple_list = []
            for x in files:
                d = os.path.getmtime(x)
                s = float(os.path.getsize(x)) / 1000.0
                file_date_tuple = (x, d, s)
                file_date_tuple_list.append(file_date_tuple)

            # Sort the tuple list by the modification time (recent changes first)
            if sortFlag:
                file_date_tuple_list.sort(key=lambda x: x[1], reverse=True)

            for fP, mT, sZ in file_date_tuple_list:
                tS = datetime.fromtimestamp(mT).strftime("%Y-%b-%d %H:%M:%S")
                rTup.append((fP, tS, sZ))
            return rTup
        except Exception as e:
            if self.__verbose:
                self.__lfh.write("+DataExchange.__getFileList() failing for patternlist %r error %r\n" % (fPatternList, str(e)))
                traceback.print_exc(file=self.__lfh)
        return rTup

    ##
    def __getArchiveFileName(self, contentType="model", formatType="pdbx", version="latest", partitionNumber="1", mileStone=None):
        (_fp, _d, f) = self.__targetFilePath(
            fileSource="archive", contentType=contentType, formatType=formatType, version=version, partitionNumber=partitionNumber, mileStone=mileStone
        )
        return f

    def __getInstanceFileName(self, contentType="model", formatType="pdbx", version="latest", partitionNumber="1", mileStone=None):
        (_fp, _d, f) = self.__targetFilePath(
            fileSource="wf-instance", contentType=contentType, formatType=formatType, version=version, partitionNumber=partitionNumber, mileStone=mileStone
        )
        return f

    def __getFilePath(self, fileSource="archive", contentType="model", formatType="pdbx", version="latest", partitionNumber="1", mileStone=None):
        (fp, _d, _f) = self.__targetFilePath(
            fileSource=fileSource, contentType=contentType, formatType=formatType, version=version, partitionNumber=partitionNumber, mileStone=mileStone
        )
        return fp

    def __targetFilePath(self, fileSource="archive", contentType="model", formatType="pdbx", version="latest", partitionNumber="1", mileStone=None):
        """ Return the file path, directory path, and filename  for the input content object if this object is valid.

            If the file path cannot be verified return None for all values
        """
        try:
            if fileSource == "session" and self.__inputSessionPath is not None:
                self.__pI.setSessionPath(self.__inputSessionPath)
            fP = self.__pI.getFilePath(
                dataSetId=self.__depDataSetId,
                wfInstanceId=self.__wfInstanceId,
                contentType=contentType,
                formatType=formatType,
                fileSource=fileSource,
                versionId=version,
                partNumber=partitionNumber,
                mileStone=mileStone,
            )
            dN, fN = os.path.split(fP)
            return fP, dN, fN
        except Exception as e:
            if self.__debug:
                self.__lfh.write(
                    "+DataExchange.__targetFilePath() failing for data set %s instance %s file source %s error %r\n"
                    % (self.__depDataSetId, self.__wfInstanceId, self.__fileSource, str(e))
                )
                traceback.print_exc(file=self.__lfh)

            return (None, None, None)



##
# File: CvsAdmin.py
# Orig: 09-April-2011  j. Westbrook
#
# Updates:
# 12-April-2011 jdw Revision checkout
# 27-Nov-2012   jdw Rename as CvsAdmin() and change logging to wwPDB style
# 27-Nov-2012   jdw Add update and commit operations
# 29-Nov-2012   jdw Refactor to separate operations for maintaining a working
#                   copy of a project (cvs sandbox) and operations
#                   which are independent of a working copy.
#  3-Dec-2012   jdw Add separate commit operations.
# 27-Jan-2013   jdw Provide return status and message text as returns for public methods
#                   in CvsSandBoxAdmin().
# 28-Jan-2013   jdw change __getCheckOutCmd() to avoid absolute path with 'co -d path'
##
"""
Wrapper class for opertations on cvs repositories.

Methods are provided to manage archiving of chemical component data files.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.001"


import sys
import os
import subprocess
import traceback
import tempfile
import shutil


class CvsWrapperBase(object):
    """Core wrapper class for opertations on cvs administrative operations on repositories.

    """

    def __init__(self, tmpPath="./", verbose=True, log=sys.stderr):
        self.__tmpPath = tmpPath
        #
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False

        self._repositoryHost = None
        self._repositoryPath = None
        self._cvsUser = None
        self._cvsPassword = None
        self._cvsRoot = None
        #
        self._wrkPath = None
        self._cvsInfoFileName = "cvsInfo.txt"
        self._cvsErrorFileName = "cvsError.txt"
        self.__outputFilePath = None
        self.__errorFilePath = None

    def setRepositoryPath(self, host, path):
        self._repositoryHost = host
        self._repositoryPath = path

    def setAuthInfo(self, user, password):
        self._cvsUser = user
        self._cvsPassword = password

    def _getRedirect(self, fileNameOut="myLog.log", fileNameErr="myLog.log", append=False):
        if append:
            if fileNameOut == fileNameErr:
                oReDir = " >> " + fileNameOut + " 2>&1 "
            else:
                oReDir = " >> " + fileNameOut + " 2>> " + fileNameErr
        else:
            if fileNameOut == fileNameErr:
                oReDir = " > " + fileNameOut + " 2>&1 "
            else:
                oReDir = " > " + fileNameOut + " 2> " + fileNameErr

        return oReDir

    def _runCvsCommand(self, myCommand):
        retcode = -100
        try:
            if self.__debug:
                self.__lfh.write("+CvsWrapperBase._runCvsCommand Command: %s\n" % myCommand)

            # retcode = subprocess.call(myCommand, stdout=self.__lfh, stderr=self.__lfh, shell=True)
            retcode = subprocess.call(myCommand, shell=True)
            if retcode != 0:
                if self.__verbose:
                    self.__lfh.write("+CvsWrapperBase.(_runCvsCommand)  Failed command: %s\n" % myCommand)
                    self.__lfh.write("+CvsWrapperBase(_runCvsCommand) Child was terminated by signal %r\n" % retcode)
                return False
            else:
                if self.__verbose:
                    self.__lfh.write("+CvsWrapperBase(_runCvsCommand) Child was terminated by signal %r\n" % retcode)
                return True
        except OSError as e:
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
                self.__lfh.write("+CvsWrapperBase(_runCvsCommand) Execution failed: %r\n" % e)
            return False

    def _setCvsRoot(self):
        try:
            self._cvsRoot = ":pserver:" + self._cvsUser + ":" + self._cvsPassword + "@" + self._repositoryHost + ":" + self._repositoryPath
            return True
        except Exception as e:
            self.__lfh.write('+CvsWrapperBase(_cvsRoot) failed')
            self.__lfh.write(e)
            return False

    def _getOutputFilePath(self):
        self.__outputFilePath = os.path.join(self._wrkPath, self._cvsInfoFileName)
        return self.__outputFilePath

    def _getErrorFilePath(self):
        self.__errorFilePath = os.path.join(self._wrkPath, self._cvsErrorFileName)
        return self.__errorFilePath

    def _getOutputText(self):
        text = ""
        try:
            filePath = self.__outputFilePath
            return self.__getTextFile(filePath)
        except Exception as e:
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
                self.__lfh.write("+CvsWrapperBase(_getOutputText) path %r %r\n" % (self.__outputFilePath, str(e)))

        return text

    def _getErrorText(self, filterInfo=False):
        text = ""
        try:
            filePath = self.__errorFilePath
            text = self.__getTextFile(filePath, filterInfo=filterInfo)
            if self.__verbose:
                self.__lfh.write("+CvsWrapperBase(_getErrorText) path: %r  text: %s\n" % (filePath, text))
                self.__lfh.flush()
        except Exception as e:
            if self.__verbose:
                traceback.print_exc(file=self.__lfh)
                self.__lfh.write("+CvsWrapperBase(_getErrorText) path %r %r\n" % (self.__errorFilePath, str(e)))
                self.__lfh.flush()

        return text

    def __getTextFile(self, filePath, filterInfo=False):  # pylint: disable=unused-argument
        text = ""
        try:
            ifh = open(filePath, "r")
            tL = []
            for line in ifh:
                if line.startswith("?"):
                    continue
                if len(str(line[:-1]).strip()) > 0:
                    tL.append(line[:-1])
            ifh.close()
            text = "\n".join(set(tL))
        except:  # noqa: E722 pylint: disable=bare-except
            pass

        return text

    def _makeTempWorkingDir(self):
        if self.__tmpPath is not None and os.path.isdir(self.__tmpPath):
            self._wrkPath = os.path.abspath(tempfile.mkdtemp("tmpdir", "tmpCVS", self.__tmpPath))

        else:
            self._wrkPath = os.path.abspath(tempfile.mkdtemp("tmpdir", "tmpCVS"))

        if self.__debug:
            self.__lfh.write("+CvsWrapperBase(_makeTempWorkingDir) Working directory path set to  %r\n" % self._wrkPath)

    def cleanup(self):
        """Cleanup any temporary files and directories created by this class.
        """
        if self._wrkPath is not None and len(self._wrkPath) > 0:
            try:
                shutil.rmtree(self._wrkPath)
                return True
            except Exception as e:
                self.__lfh.write('cleanup - unable to remove self._wrkpath')
                self.__lfh.write(e)
                return False
        else:
            return True


class CvsAdmin(CvsWrapperBase):
    """Wrapper class for opertations on cvs administrative operations on repositories.
    """

    def __init__(self, tmpPath="./", verbose=True, log=sys.stderr):
        super(CvsAdmin, self).__init__(tmpPath=tmpPath, verbose=verbose, log=log)
        self.__tmpPath = tmpPath
        #
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False

    def getHistory(self, cvsPath):
        """ Return the history text for project files identified by cvsPath in the
            current repository.
        """
        text = ""
        ok = False
        cmd = self.__getHistoryCmd(cvsPath)
        if cmd is not None:
            ok = self._runCvsCommand(myCommand=cmd)
            text = self._getOutputText()
        else:
            text = "History command failed with repository command processing error"

        return (ok, text)

    def getRevisionList(self, cvsPath):
        """ Return a list of tuples containing the revision identifiers for the input file.

            Return data has the for [(RevId, A/M, timeStamp),...] where A=Added and M=Modified.
        """
        revList = []
        ok = False
        cmd = self.__getHistoryCmd(cvsPath)
        if cmd is not None:
            ok = self._runCvsCommand(myCommand=cmd)
            revList = self.__extractRevisions()
        else:
            _text = "Revision history command failed with repository command processing error"  # noqa: F841

        return (ok, revList)

    def checkOutFile(self, cvsPath, outPath, revId=None):
        """ Perform CVS checkout operation for the project files identified by the input cvsPath
            subject to the input revision identifier.

            Files that are checked out are then subsequently copied to outPath.

            Note that outPath will not be a CVS working copy (sandbox) after this operation.
        """
        text = ""
        ok = False
        (pth, fn) = os.path.split(cvsPath)
        if self.__verbose:
            self.__lfh.write("+CvsAdmin(checkOutFile) Cvs directory %s   target file name %s\n" % (pth, fn))
        if len(fn) > 0:
            cmd = self.__getCheckOutCmd(cvsPath, outPath, revId)
            if cmd is not None:
                ok = self._runCvsCommand(myCommand=cmd)
                text = self._getErrorText()
            else:
                text = "Check out failed with repository command processing error"
        else:
            self.__lfh.write("+ERROR - CvsAdmin(checkOutFile) cannot check out project path %s\n" % cvsPath)
            text = "Check out failed with repository project path issue: %s" % cvsPath

        return (ok, text)

    def __getHistoryCmd(self, cvsPath):
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        outPath = self._getOutputFilePath()
        errPath = self._getErrorFilePath()
        if self._setCvsRoot():
            cmd = "cvs -d " + self._cvsRoot + " history -a -x AM " + cvsPath + self._getRedirect(fileNameOut=outPath, fileNameErr=errPath)
        else:
            cmd = None
        return cmd

    def __getCheckOutCmd(self, cvsPath, outPath, revId=None):
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        errPath = self._getErrorFilePath()
        (pth, fn) = os.path.split(cvsPath)
        if self.__verbose:
            self.__lfh.write("+CvsAdmin(__getCheckOutCmd) CVS directory %s  target file name %s\n" % (pth, fn))
        lclPath = os.path.join(self._wrkPath, cvsPath)
        outPathAbs = os.path.abspath(outPath)
        #
        #
        if self._setCvsRoot():
            if revId is None:
                rS = " "
                cmd = (
                    "cd "
                    + self._wrkPath
                    + " ; cvs -d "
                    + self._cvsRoot
                    + " co "
                    + cvsPath
                    + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath)
                    + " ; "
                    + " mv -f  "
                    + lclPath
                    + " "
                    + outPathAbs
                    + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True)
                    + " ; "
                )
                # cmd="cvs -d " +  self._cvsRoot  + " co -d " + self._wrkPath + " " + cvsPath +  \
                #     self._getRedirect(fileNameOut=errPath,fileNameErr=errPath) + ' ; '  + \
                #     " mv -f  " + lclPath + " " + outPath + self._getRedirect(fileNameOut=errPath,fileNameErr=errPath,append=True) + ' ; '
            else:
                rS = " -r " + revId + " "
                cmd = (
                    "cd "
                    + self._wrkPath
                    + "; cvs -d "
                    + self._cvsRoot
                    + " co "
                    + rS
                    + " "
                    + cvsPath
                    + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath)
                    + " ; "
                    + " mv -f  "
                    + lclPath
                    + " "
                    + outPathAbs
                    + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True)
                    + " ; "
                )
                # cmd="cvs -d " +  self._cvsRoot  + " co -d " + self._wrkPath + rS +  " " + cvsPath + \
                #     self._getRedirect(fileNameOut=errPath,fileNameErr=errPath) + ' ; '  + \
                #     " mv -f  " + lclPath + " " + outPath + self._getRedirect(fileNameOut=errPath,fileNameErr=errPath,append=True) + ' ; '
        else:
            cmd = None
        return cmd

    def __extractRevisions(self):
        """Extract revisions details from the last history command.
        """
        revList = []
        try:
            fName = self._getOutputFilePath()
            if self.__verbose:
                self.__lfh.write("+CvsAdmin(__extraRevisions) Reading revisions from %r\n" % fName)
            ifh = open(fName, "r")
            for line in ifh.readlines():
                fields = line[:-1].split()
                typeCode = str(fields[0])
                revId = str(fields[5])
                timeStamp = str(fields[1] + ":" + fields[2])
                revList.append((revId, typeCode, timeStamp))
        except Exception as e:
            if self.__verbose:
                self.__lfh.write("+CvsAdmin(__extraRevisions) Extracting revision list for : %s %s\n" % (fName, str(e)))

        revList.reverse()
        self.__lfh.write("Ordered revision list %r\n" % revList)

        return revList


class CvsSandBoxAdmin(CvsWrapperBase):
    """Wrapper class for opertations on cvs working directories (aka cvs sandboxes).

    """

    def __init__(self, tmpPath="./", verbose=True, log=sys.stderr):
        super(CvsSandBoxAdmin, self).__init__(tmpPath=tmpPath, verbose=verbose, log=log)
        self.__tmpPath = tmpPath
        #
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False
        #
        self.__sandBoxTopPath = None

    def setSandBoxTopPath(self, dirPath):
        """ Assign the path that contains or will contain the working copy of the cvs project.

            Must be an existing writable path.

        """
        if os.access(dirPath, os.W_OK):
            self.__sandBoxTopPath = dirPath
            return True
        else:
            if not os.path.exists(dirPath):
                try:
                    os.makedirs(dirPath)
                    if os.access(dirPath, os.W_OK):
                        self.__sandBoxTopPath = dirPath
                        return True
                except Exception as e:
                    self.__lfh.write("+CvsAdmin failed to make CVS top path")
                    self.__lfh.write(e)
            return False

    def getSandBoxTopPath(self):
        return os.path.abspath(self.__sandBoxTopPath)

    def checkOut(self, projectPath=None, revId=None):
        """ Create CVS sandbox working copy of the input project path within the current repository.

        """
        if self.__verbose:
            self.__lfh.write("\n+CvsSandBoxAdmin(checkOut) Checking out CVS repository working path %s project file path %s\n" % (self.__sandBoxTopPath, projectPath))
        text = ""
        ok = False
        if self.__sandBoxTopPath is not None and projectPath is not None:
            cmd = self.__getCheckOutProjectCmd(projectPath, revId=revId)
            if self.__verbose:
                self.__lfh.write("\n+CvsSandBoxAdmin(checkOut) checkout command %s\n" % cmd)
            if cmd is not None:
                ok = self._runCvsCommand(myCommand=cmd)
                text = self._getErrorText()
            else:
                text = "Check out failed with repository command processing error"
        else:
            self.__lfh.write("+ERROR - CvsSandBoxAdmin(checkOut) cannot check out project path %s\n" % projectPath)
            text = "Check out failed with repository project path issue: %s" % projectPath

        return (ok, text)

    def updateList(self, dataList, procName, optionsD, workingDir):  # pylint: disable=unused-argument
        """  Implements an interface for multiprocessing module --

             input is [(CvsProjectDir, relativePath, pruneFlag),...]

             returns -  successList,resultList=successList,diagList
        """
        retList = []
        diagTextList = []
        for dTup in dataList:
            pDir, relPath, _prune = dTup
            ok, text = self.update(projectDir=pDir, relProjectPath=relPath, prune=True, fetchErrorLog=True, appendErrors=True)
            diagTextList.append(text)
            if self.__verbose:
                self.__lfh.write("+CvsSandBoxAdmin(updateList) process %s project %s path %s status %r diagnostics:\n%s\n" % (procName, pDir, relPath, ok, text))
                self.__lfh.write("\n+CvsSandBoxAdmin(updateList) current diagnostics length %d\n" % len(diagTextList))
                self.__lfh.flush()
            if ok:
                retList.append(dTup)
        return retList, retList, diagTextList

    def update(self, projectDir, relProjectPath=".", prune=False, fetchErrorLog=True, appendErrors=False):
        """ Update CVS sandbox working copy of the input project path.   The project path must
            correspond to an existing working copy of the repository.

        """
        if self.__verbose:
            self.__lfh.write(
                "\n+CvsSandBoxAdmin(update) Updating CVS repository working path %s project %s relative file path %s\n" % (self.__sandBoxTopPath, projectDir, relProjectPath)
            )
        targetPath = os.path.join(self.__sandBoxTopPath, projectDir)
        text = ""
        ok = False
        if os.access(targetPath, os.W_OK):
            cmd = self.__getUpdateCmd(projectDir, relProjectPath=relProjectPath, prune=prune, appendErrors=appendErrors)
            if self.__debug:
                self.__lfh.write("\n+CvsSandBoxAdmin(update) update command %s\n" % cmd)
            if cmd is not None:
                ok = self._runCvsCommand(myCommand=cmd)
                if fetchErrorLog:
                    text = self._getErrorText(filterInfo=True)
                else:
                    text = "+CvsSandBoxAdmin(update) failing update command %s" % cmd
            else:
                text = "Update failed with repository command processing error"
        else:
            self.__lfh.write("+ERROR - CvsSandBoxAdmin(update) top sandbox path %s project dir %s \n" % (self.__sandBoxTopPath, projectDir))
            if not os.path.exists(targetPath):
                try:
                    os.makedirs(targetPath)
                except Exception as e:
                    self.__lfh.write("+ERROR - CvsSandBoxAdmin(update) cannot make project path %s\n" % targetPath)
                    self.__lfh.write(e)
            if os.access(self.__sandBoxTopPath, os.W_OK):
                # try a full checkout --
                #
                ok, text = self.checkOut(projectPath=projectDir, revId=None)
            else:
                text = "Update failed with repository project path issue: %s" % targetPath
                self.__lfh.write("+ERROR - CvsSandBoxAdmin(update) cannot update project path %s\n" % targetPath)

        return (ok, text)

    def add(self, projectDir, relProjectPath):
        """ Add an new definition in CVS working direcotry in the input project path.   The project path must
            correspond to an existing file path in the local working copy.

        """
        if self.__verbose:
            self.__lfh.write("\n+CvsSandBoxAdmin(add) Add %s to project %s in CVS repository working path %s\n" % (relProjectPath, projectDir, self.__sandBoxTopPath))
        targetPath = os.path.join(self.__sandBoxTopPath, projectDir, relProjectPath)
        text = ""
        ok = False
        if os.access(targetPath, os.W_OK):
            cmd = self.__getAddCommitCmd(projectDir, relProjectPath)
            if cmd is not None:
                ok = self._runCvsCommand(myCommand=cmd)
                text = self._getErrorText()
            else:
                text = "Add failed with repository command processing error"
        else:
            text = "Add failed due with repository project path issue: %s" % targetPath
            self.__lfh.write("+ERROR - CvsSandBoxAdmin(add) cannot add project path %s\n" % targetPath)

        return (ok, text)

    def commit(self, projectDir, relProjectPath):
        """ Commit changes in the input project/file path to the CVS repository. The project path must
            correspond to an existing path in the local working copy.

        """
        if self.__verbose:
            self.__lfh.write(
                "\n+CvsSandBoxAdmin(commit) Commit changes to %s in project %s in CVS repository working path %s\n" % (relProjectPath, projectDir, self.__sandBoxTopPath)
            )
        targetPath = os.path.join(self.__sandBoxTopPath, projectDir, relProjectPath)
        text = ""
        ok = False
        if os.access(targetPath, os.W_OK):
            cmd = self.__getCommitCmd(projectDir, relProjectPath)
            if cmd is not None:
                ok = self._runCvsCommand(myCommand=cmd)
                text = self._getErrorText()
            else:
                text = "Commit failed with repository command processing error"
        else:
            text = "Commit failed due with repository project path issue: %s" % targetPath
            self.__lfh.write("+ERROR - CvsSandBoxAdmin(commit) cannot commit project path %s\n" % targetPath)

        return (ok, text)

    def remove(self, projectDir, relProjectPath, saveCopy=True):
        """ Remove from the CVS sandbox working copy the input project path.   The project path must
            correspond to an existing path in the local working copy.

            if saveCopy=True then preserve a copy of the remove target in the reserved REMOVED path
            of the repository.
        """
        if self.__verbose:
            self.__lfh.write("\n+CvsSandBoxAdmin(remove) Remove %s from project %s in CVS repository working path %s\n" % (relProjectPath, projectDir, self.__sandBoxTopPath))

        text = ""
        ok = False
        if (relProjectPath is None) or (len(relProjectPath) < 3):
            return (ok, text)
        #
        targetPath = os.path.join(self.__sandBoxTopPath, projectDir, relProjectPath)

        if self.__verbose:
            self.__lfh.write("\n+CvsSandBoxAdmin(remove) Remove target file path is %s\n" % targetPath)

        if os.access(targetPath, os.W_OK):
            #
            if saveCopy:
                (_pth, fn) = os.path.split(relProjectPath)
                savePath = os.path.join(self.__sandBoxTopPath, projectDir, "REMOVED", fn)

                # relSavePath = os.path.join("REMOVED", fn)
                shutil.copyfile(targetPath, savePath)
                # (ok1,saveText)=self.add(projectDir,relSavePath)
                # if not ok1:
                #    return (ok1,saveText)
            cmd = self.__getRemoveCommitCmd(projectDir, relProjectPath)
            if cmd is not None:
                ok = self._runCvsCommand(myCommand=cmd)
                text = self._getErrorText()
            else:
                text = "Remove failed with repository command processing error"
        else:
            text = "Remove failed due with repository project path issue: %s" % targetPath
            self.__lfh.write("+ERROR - CvsSandBoxAdmin(remove) cannot remove project path %s\n" % targetPath)

        return (ok, text)

    def removeDir(self, projectDir, relProjectPath):
        """ Remove from the CVS sandbox working directory the input empty directory.

        """
        if self.__verbose:
            self.__lfh.write("\n+CvsSandBoxAdmin(removeDir) Remove %s from project %s in CVS repository working path %s\n" % (relProjectPath, projectDir, self.__sandBoxTopPath))

        text = ""
        ok = False
        if (relProjectPath is None) or (len(relProjectPath) < 3):
            return (ok, text)
        #
        targetPath = os.path.join(self.__sandBoxTopPath, projectDir, relProjectPath)

        if os.access(targetPath, os.W_OK):
            #
            cmd = self.__getRemoveDirCommitCmd(projectDir, relProjectPath)
            if cmd is not None:
                ok = self._runCvsCommand(myCommand=cmd)
                text = self._getErrorText()
            else:
                text = "Remove directory failed with repository command processing error"
        else:
            text = "Remove directory failed due with repository project path issue: %s" % targetPath
            self.__lfh.write("+ERROR - CvsSandBoxAdmin(remove) cannot remove project path %s\n" % targetPath)

        return (ok, text)

    def __getCheckOutProjectCmd(self, relProjectPath, revId=None):
        """ Return CVS command for checkout of a complete project from the current repository.
        """
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        errPath = self._getErrorFilePath()
        if self._setCvsRoot():
            cmd = " cd " + self.__sandBoxTopPath + " ; "
            if revId is None:
                cmd += "cvs -d " + self._cvsRoot + " co " + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath) + " ; "
            else:
                rS = " -r " + revId + " "
                cmd += "cvs -d " + self._cvsRoot + " co " + rS + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath) + " ; "
        else:
            cmd = None
        return cmd

    def __getUpdateProjectCmd(self, projectDir, prune=False, appendErrors=False):
        """ Return CVS command for updating a complete project working directory from current repository.
        """
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        errPath = self._getErrorFilePath()
        if self._setCvsRoot():
            if prune:
                pF = " -P "
            else:
                pF = " "
            targetPath = os.path.join(self.__sandBoxTopPath, projectDir)
            cmd = " cd " + targetPath + "; "
            cmd += "cvs -d " + self._cvsRoot + " update -C -d " + pF + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=appendErrors) + " ; "
        else:
            cmd = None
        return cmd

    def __getUpdateCmd(self, projectDir, relProjectPath, prune=False, appendErrors=False):
        """ Return CVS command for updating the input relative path within project working
            directory from current repository.
        """
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        errPath = self._getErrorFilePath()
        if self._setCvsRoot():
            if prune:
                pF = " -P "
            else:
                pF = " "
            targetPath = os.path.join(self.__sandBoxTopPath, projectDir)
            cmd = " cd " + targetPath + "; "
            cmd += "cvs -q -d " + self._cvsRoot + " update -C -d " + pF + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=appendErrors) + " ; "
        else:
            cmd = None
        return cmd

    def __getAddCommitCmd(self, projectDir, relProjectPath, message="Initial version"):
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        errPath = self._getErrorFilePath()
        if self._setCvsRoot():
            cmd = " cd " + os.path.join(self.__sandBoxTopPath, projectDir) + " ; "
            if message is not None and len(message) > 0:
                qm = ' -m "' + message + '" '
            cmd += "cvs -d " + self._cvsRoot + " add " + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath) + " ; "
            cmd += "cvs -d " + self._cvsRoot + " commit " + qm + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True) + " ; "
        else:
            cmd = None
        return cmd

    def __getCommitCmd(self, projectDir, relProjectPath, message="Automated update"):
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        errPath = self._getErrorFilePath()
        if self._setCvsRoot():
            cmd = " cd " + os.path.join(self.__sandBoxTopPath, projectDir) + " ; "
            if message is not None and len(message) > 0:
                qm = ' -m "' + message + '" '
            cmd += "cvs -d " + self._cvsRoot + " commit " + qm + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True) + " ; "
        else:
            cmd = None
        return cmd

    def __getRemoveCommitCmd(self, projectDir, relProjectPath, message="File removed"):
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        errPath = self._getErrorFilePath()
        if self._setCvsRoot():
            cmd = " cd " + os.path.join(self.__sandBoxTopPath, projectDir) + " ; "
            if message is not None and len(message) > 0:
                qm = ' -m "' + message + '" '
            cmd += "cvs -d " + self._cvsRoot + " remove -f " + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath) + " ; "
            cmd += "cvs -d " + self._cvsRoot + " commit " + qm + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True) + " ; "
        else:
            cmd = None
        return cmd

    def __getRemoveDirCommitCmd(self, projectDir, relProjectPath, message="Directory removed"):
        if self._wrkPath is None:
            self._makeTempWorkingDir()
        errPath = self._getErrorFilePath()
        if self._setCvsRoot():
            cmd = " cd " + os.path.join(self.__sandBoxTopPath, projectDir) + " ; "
            if message is not None and len(message) > 0:
                qm = ' -m "' + message + '" '
            cmd += "cvs -d " + self._cvsRoot + " remove " + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath) + " ; "
            cmd += "cvs -d " + self._cvsRoot + " commit " + qm + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True) + " ; "
            cmd += " rm -rf " + relProjectPath + self._getRedirect(fileNameOut=errPath, fileNameErr=errPath, append=True) + " ; "
        else:
            cmd = None
        return cmd



##
# File: GraphicsContext3D.py
# Date: 24-June-2012
#
# Updates:
# 27-June-2012 jdw Change magnification on zoomto and quoting policy on labels.
# 28-Sept-2012 jdw Adjust the Jmol setup to improve the depiction.
#
##
"""
Construct a 3D graphics context from selected rows in PDBx/mmCIF data catagories.


"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"


import sys
import traceback

from mmcif_utils.persist.PdbxPersist import PdbxPersist


class GraphicsContext3D(object):
    """ Construct a 3D graphics context from selected rows in PDBx/mmCIF data catagories.

    Only app3D='JMol' is currently supported.
    """

    def __init__(self, app3D="JMol", verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__app3D = app3D
        self.__persistFilePath = None

        self.__setup()
        #

    def __setup(self):
        """ Category dictionary containing data attribute names which define 3D context.

        For each category with a graphics context a list of atom/component/polymer features is
        is provided.  The feature are defined as a set of standard keys defining the
        the author nomenclature for the feature.

        """
        self.__jMolDefaultSelection = "select all; label off; "
        # self.__jMolSetup=" background black; wireframe only; wireframe 0.15; labels off; slab 100; depth 40; slab on; "

        self.__jMolSetup = (
            " background black; font label 20 monospaced bold; set fontScaling OFF; color label white; wireframe only; wireframe 0.10; labels off; slab 70; depth 25; slab on; "
        )
        #
        # Category types to customize context assignments -
        #
        self.__searchContextCategoryList = ["struct_site"]
        self.__rangeTypeCategoryList = ["struct_conf", "struct_sheet_range"]
        self.__atomContextCategoryList = ["atom_site", "struct_conn", "pdbx_struct_sheet_hbond"]
        self.__componentContextCategoryList = ["struct_conf", "struct_sheet_range", "pdbx_poly_seq_scheme", "pdbx_nonpoly_scheme"]
        self.__polymerContextCategoryList = []
        #
        #  Templates for extracting structur features from selected categories.
        #
        self.__d = {
            "struct_conn": [
                {
                    "auth_asym_id": "ptnr1_auth_asym_id",
                    "auth_comp_id": "ptnr1_auth_comp_id",
                    "auth_seq_id": "ptnr1_auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "pdbx_ptnr1_PDB_ins_code",
                    "atom_id": "ptnr1_label_atom_id",
                    "sym_op": "ptnr1_symmetry",
                    "model_num": None,
                },
                {
                    "auth_asym_id": "ptnr2_auth_asym_id",
                    "auth_comp_id": "ptnr2_auth_comp_id",
                    "auth_seq_id": "ptnr2_auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "pdbx_ptnr2_PDB_ins_code",
                    "atom_id": "ptnr2_label_atom_id",
                    "sym_op": "ptnr2_symmetry",
                    "model_num": None,
                },
            ],
            "struct_conf": [
                {
                    "auth_asym_id": "beg_auth_asym_id",
                    "auth_comp_id": "beg_auth_comp_id",
                    "auth_seq_id": "beg_auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "pdbx_beg_PDB_ins_code",
                    "atom_id": "auth_atom_id",
                    "sym_op": None,
                    "model_num": None,
                },
                {
                    "auth_asym_id": "end_auth_asym_id",
                    "auth_comp_id": "end_auth_comp_id",
                    "auth_seq_id": "end_auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "end_beg_PDB_ins_code",
                    "atom_id": "auth_atom_id",
                    "sym_op": None,
                    "model_num": None,
                },
            ],
            "struct_site_gen": [
                {
                    "auth_asym_id": "auth_asym_id",
                    "auth_comp_id": "auth_comp_id",
                    "auth_seq_id": "auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "pdbx_auth_ins_code",
                    "atom_id": "auth_atom_id",
                    "sym_op": "symmetry",
                    "model_num": None,
                }
            ],
            "atom_site": [
                {
                    "auth_asym_id": "auth_asym_id",
                    "auth_comp_id": "auth_comp_id",
                    "auth_seq_id": "auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "pdbx_PDB_ins_code",
                    "atom_id": "auth_atom_id",
                    "sym_op": None,
                    "model_num": "pdbx_PDB_model_num",
                }
            ],
            "pdbx_poly_seq_scheme": [
                {
                    "auth_asym_id": "pdb_strand_id",
                    "auth_comp_id": "pdb_mon_id",
                    "auth_seq_id": "pdb_seq_num",
                    "auth_seq_id_range": None,
                    # 'auth_comp_id': 'auth_mon_id',
                    # 'auth_seq_id' : 'auth_seq_num',
                    "ins_code": "pdb_ins_code",
                    "atom_id": "auth_atom_id",
                    "sym_op": None,
                    "model_num": None,
                }
            ],
            "pdbx_nonpoly_scheme": [
                {
                    "auth_asym_id": "pdb_strand_id",
                    "auth_comp_id": "pdb_mon_id",
                    "auth_seq_id": "pdb_seq_num",
                    "auth_seq_id_range": None,
                    # 'auth_comp_id': 'auth_mon_id',
                    # 'auth_seq_id' : 'auth_seq_num',
                    "ins_code": "pdb_ins_code",
                    "atom_id": "auth_atom_id",
                    "sym_op": None,
                    "model_num": None,
                }
            ],
            "struct_sheet_range": [
                {
                    "auth_asym_id": "beg_auth_asym_id",
                    "auth_comp_id": "beg_auth_comp_id",
                    "auth_seq_id": "beg_auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "pdbx_beg_PDB_ins_code",
                    "atom_id": None,
                    "sym_op": "symmetry",
                    "model_num": None,
                },
                {
                    "auth_asym_id": "end_auth_asym_id",
                    "auth_comp_id": "end_auth_comp_id",
                    "auth_seq_id": "end_auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "pdbx_end_PDB_ins_code",
                    "atom_id": None,
                    "sym_op": "symmetry",
                    "model_num": None,
                },
            ],
            "pdbx_struct_sheet_hbond": [
                {
                    "auth_asym_id": "range_1_auth_asym_id",
                    "auth_comp_id": "range_1_auth_comp_id",
                    "auth_seq_id": "range_1_auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "range_1_PDB_ins_code",
                    "atom_id": "range_1_auth_atom_id",
                    "sym_op": None,
                    "model_num": None,
                },
                {
                    "auth_asym_id": "range_2_auth_asym_id",
                    "auth_comp_id": "range_2_auth_comp_id",
                    "auth_seq_id": "range_2_auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "range_2_PDB_ins_code",
                    "atom_id": "range_2_auth_atom_id",
                    "sym_op": None,
                    "model_num": None,
                },
            ],
            "pdbx_distant_solvent_atoms": [
                {
                    "auth_asym_id": "auth_asym_id",
                    "auth_comp_id": "auth_comp_id",
                    "auth_seq_id": "auth_seq_id",
                    "auth_seq_id_range": None,
                    "ins_code": "PDB_ins_code",
                    "atom_id": "auth_atom_id",
                    "sym_op": None,
                    "model_num": "PDB_model_num",
                }
            ],
        }

    #
    #  Public methods
    #
    def getCategoriesWithContext(self):
        """  Return the list of categories with a defined graphics context.
        """
        cL = []
        cL.extend(self.__d.keys())
        cL.extend(self.__searchContextCategoryList)
        return cL

    def setPersistStorePath(self, persistFilePath):
        """  Set the path of persistent store.

        Required for categories in the __searchContextCategoryList.

        """
        self.__persistFilePath = persistFilePath

    def getGraphicsContext(self, categoryName=None, rowDictList=None):
        """ Create a command string to highlight the 3D graphics context for the
            input list of rows (stored with attribute keys) in the target category.

            Return a script command appropriate for the current 3D graphics application.
        """
        #
        if categoryName in self.__searchContextCategoryList:
            return self.__getContextViaSearch(categoryName=categoryName, rowDictList=rowDictList)
        #
        contextList = []
        for rowDict in rowDictList:
            cS = self.__createContext(categoryName=categoryName, rowDict=rowDict)
            if (cS is not None) and (len(cS) > 0):
                contextList.extend(cS)

        rL = []
        rL.append(self.__jMolSetup)
        rL.append(self.__jMolDefaultSelection)
        #
        if len(contextList) > 0:
            tS = "select ( " + ",".join(contextList) + " ) ; "
            rL.append(tS)
            lS = self.__assignLabelStyle(categoryName=categoryName)
            rL.append(lS)
            rL.append("zoomto (selected) 900; ")
        #
        return "".join(rL)

    #
    # Internal methods for here on --
    #
    def __createContext(self, categoryName=None, rowDict=None):
        """  Wrapper to create the feature selection ---
        """
        contextL = []

        if categoryName in self.__rangeTypeCategoryList:
            contextL = self.__createComponentRangeContext(categoryName=categoryName, rowDict=rowDict)
        elif categoryName in self.__d.keys():
            contextL = self.__createContextSimple(categoryName=categoryName, rowDict=rowDict)
        else:
            pass

        return contextL

    def __getContextViaSearch(self, categoryName=None, rowDictList=None):
        """ Establish the context for cases where the structural details may be determined from
        related data categories.
        """
        #
        gcS = self.__jMolDefaultSelection
        if categoryName == "struct_site":
            searchKeyName = "site_id"
            searchCategoryName = "struct_site_gen"
            #
            rDL = []
            for rowDict in rowDictList:
                keyValue = self.__getStringValue(rowDict)
                rDL.extend(self.__searchAttribute(keyValue=keyValue, searchKeyName=searchKeyName, searchCategoryName=searchCategoryName))
            #
            #
            gcS = self.getGraphicsContext(categoryName=searchCategoryName, rowDictList=rDL)
        elif categoryName == "struct_sheet":
            searchKeyName = "sheet_id"
            searchCategoryName = "struct_sheet_range"
            #
            rDL = []
            for rowDict in rowDictList:
                keyValue = self.__getStringValue(rowDict)
                rDL.extend(self.__searchAttribute(keyValue=keyValue, searchKeyName=searchKeyName, searchCategoryName=searchCategoryName))
            #
            #
            gcS = self.getGraphicsContext(categoryName=searchCategoryName, rowDictList=rDL)
        else:
            pass

        return gcS

    def __createContextSimple(self, categoryName=None, rowDict=None):
        """  Create a graphics context from the list of feature templates for this category.

             The features are treated independently and the associated contexts are returned as
             a list of JMol "atom expressions" appropriate for a JMol "select" statement.
        """
        contextL = []
        catNameLC = categoryName.lower()
        if catNameLC in self.__d:
            for fD in self.__d[catNameLC]:
                fI = self.__extractValues(fD, rowDict)
                fS = self.__assignFeatureContext(self.__app3D, fI)
                if self.__debug:
                    self.__lfh.write("+GraphicsContext3D.__createContext - feature dict     %r\n" % fD.items())
                    self.__lfh.write("+GraphicsContext3D.__createContext - row     dict     %r\n" % rowDict.items())
                    self.__lfh.write("+GraphicsContext3D.__createContext - feature instance %r\n" % fI.items())
                    self.__lfh.write("+GraphicsContext3D.__createContext - feature context  %s\n" % fS)

                if fS is not None and len(fS) > 0:
                    contextL.append(fS)
        else:
            pass
        #
        return contextL

    def __createComponentRangeContext(self, categoryName=None, rowDict=None):
        """  Create a "component range" graphics context from a pair of features templates for this category.

             The feature pair are treated as defining a contiguous range of components and a range style
             JMol "atom expression" appropriate for a JMol "select" statement is returned.
        """

        contextL = []
        catNameLC = categoryName.lower()
        if (catNameLC in self.__d) and (len(self.__d[catNameLC]) == 2):
            fD1 = self.__d[catNameLC][0]
            fD2 = self.__d[catNameLC][1]
            fI1 = self.__extractValues(fD1, rowDict)
            fI2 = self.__extractValues(fD2, rowDict)
            if self.__debug:
                self.__lfh.write("+GraphicsContext3D.__createContext - feature instance 1 %r\n" % fI1.items())
                self.__lfh.write("+GraphicsContext3D.__createContext - feature instance 2 %r\n" % fI2.items())
            #
            # Check if the features are compatible for a range context -
            #
            if (
                (fI1["auth_asym_id"] == fI2["auth_asym_id"])
                and (fI1["model_num"] == fI2["model_num"])
                and (fI1["sym_op"] == fI2["sym_op"])
                and (fI1["auth_seq_id"] is not None)
                and (fI2["auth_seq_id"] is not None)
            ):
                #
                # create the component range assignment in the first instance.
                fI1["auth_seq_id_range"] = (fI1["auth_seq_id"], fI2["auth_seq_id"])
                fI1["auth_seq_id"] = None
                fS = self.__assignFeatureContext(self.__app3D, fI1)
                if fS is not None and len(fS) > 0:
                    contextL.append(fS)
            else:
                # Default to range end points -
                fS = self.__assignFeatureContext(self.__app3D, fI1)
                if fS is not None and len(fS) > 0:
                    contextL.append(fS)
                fS = self.__assignFeatureContext(self.__app3D, fI2)
                if fS is not None and len(fS) > 0:
                    contextL.append(fS)
        else:
            pass
        #
        return contextL

    def __assignFeatureContext(self, app3D, featureInstDict=None):
        """ The general syntax for JMol is  [<compId>]<seqId>^<insertCode>:<authAsymId>.<atomId>/<model_num>
                                            [<compId>]<beg_seqId>-<end_seqId>^<insertCode>:<authAsymId>.<atomId>/<model_num>
        """
        if featureInstDict is None or featureInstDict == {}:
            return ""
        sL = []
        if app3D == "JMol":
            if featureInstDict["sym_op"] is not None and featureInstDict["sym_op"] != "1_555":
                # bail out here if the feature is not in the deposited coordinates.
                return
            if featureInstDict["auth_comp_id"] is not None:
                s = "[%s]" % featureInstDict["auth_comp_id"]
                sL.append(s)
            if featureInstDict["auth_seq_id"] is not None:
                s = "%s" % featureInstDict["auth_seq_id"]
                sL.append(s)
            if featureInstDict["auth_seq_id_range"] is not None:
                s = "%s-%s" % featureInstDict["auth_seq_id_range"]
                sL.append(s)
            if featureInstDict["ins_code"] is not None:
                s = "^%s" % featureInstDict["ins_code"]
                sL.append(s)
            if featureInstDict["auth_asym_id"] is not None:
                s = ":%s" % featureInstDict["auth_asym_id"]
                sL.append(s)
            if featureInstDict["atom_id"] is not None:
                s = ".%s" % featureInstDict["atom_id"]
                sL.append(s)
            if featureInstDict["model_num"] is not None:
                s = "/%s" % featureInstDict["model_num"]
                sL.append(s)
        else:
            sL = []
        return "".join(sL)

    def __getFirstObject(self, persistFilePath, objectName=None):
        """ Open the persistent data store and fetch the input object name from the first container.

        Note -- Will be used for more complex cases which require additional information from
        coordinate model file to resolve the graphics context.
        """
        #
        try:
            myPersist = PdbxPersist(self.__verbose, self.__lfh)
            indexD = myPersist.getIndex(dbFileName=persistFilePath)
            (firstContainerName, _type) = indexD["__containers__"][0]

            if self.__debug:
                self.__lfh.write("GraphicsContext3D.getFirstObject() container name list %r\n" % indexD.items())
            # myPersist.open(dbFileName=persistFilePath,flag='r')
            # myPersist.recover()
            # cList=myPersist.getContainerNameList()
            # self.__lfh.write("GraphicsContext3D.getFirstObject() container name list %r\n" % cList)

            myObj = myPersist.fetchOneObject(dbFileName=persistFilePath, containerName=firstContainerName, objectName=objectName)
            return myObj
        except Exception as e:
            if self.__verbose:
                self.__lfh.write("+ERROR- GraphicsContext3D.getFirstObject() Read failed for file %s %s\n" % (persistFilePath, str(e)))
                traceback.print_exc(file=self.__lfh)
            return None

    def __getStringValue(self, rowDict):
        if "id" in rowDict and len(rowDict["id"]) > 0 and rowDict["id"] != "?" and rowDict["id"] != ".":
            return rowDict["id"]
        else:
            return None

    def __searchAttribute(self, keyValue=None, searchKeyName=None, searchCategoryName=None):
        """ Search input category for rows where the attribue searchKeyName equals
            the input keyValue.

            Return a list of rows stored as dictionaries with attribute names mapped to  values.
        """
        rDL = []
        if keyValue is not None:
            myCatObj = self.__getFirstObject(persistFilePath=self.__persistFilePath, objectName=searchCategoryName)
            if myCatObj is not None:
                aL = myCatObj.getAttributeList()
                indexSearchKey = aL.index(searchKeyName)
                #
                # assemble a row dictionary list to hold search results.
                rDL = []
                for row in myCatObj.getRowList():
                    if keyValue == row[indexSearchKey]:
                        rD = {}
                        for ii, rVal in enumerate(row):
                            rD[aL[ii]] = rVal
                        # self.__lfh.write("Row dictionary: %r\n" % rD.items())
                        rDL.append(rD)
                        #
                        if self.__debug:
                            self.__lfh.write("GraphicsContext3D.__getContextViaSearch KeyValue %s row : %r\n" % (keyValue, rD.items()))
        return rDL

    def __assignLabelStyle(self, categoryName=None):
        if categoryName in self.__atomContextCategoryList:
            return "label '%c:%n:%r:%a' ; "
        elif categoryName in self.__componentContextCategoryList:
            return "label '%c:%n:%r' ; "
        elif categoryName in self.__polymerContextCategoryList:
            return "label '%c' ; "
        else:
            return "label '%c:%n:%r' ; "

    def __extractValues(self, attribDict, rowDict):
        """  Return a value dictionary
        """
        rD = {}
        for ky, attrib in attribDict.items():
            rD[ky] = None
            if attrib is not None and attrib in rowDict:
                if (len(rowDict[attrib]) > 0) and (rowDict[attrib] != "?") and (rowDict[attrib] != "."):
                    rD[ky] = rowDict[attrib]
        return rD



##
# File:    DataReference.py
# Date:    30-Mar-2010
#
#
# Updates:
#  17-Apr-2010 - jdw add isReferenceValid()
#  29-Apr-2010 - jdw add support for director references.
#  13-Sep-2010 - jdw added getFilePathExists()
#  29-Jun-2011 - jdw fix version sorting issue
#  25-Feb-2013   jdw provided defaulted siteId/verbose/log in DataFileReference constructor
#  26-Feb-2013   jdw added session/wf-session storage type.
#  02-Mar-2013   jdw add class ReferenceFileInfo()
#  02-Mar-2013   jdw added class ReferenceFileComponents()
#  05-Mar-2013   jdw add initialization for session data set id.
#  07-Mar-2013   jdw make sesssion data set ids always upper case.
#  22-Mar-2013   jdw make files in workflow instance storage use the deposition data set id as
#                    leading file name element
#  25-Mar-2013   jdw make session data set id distinct from archive/workflow depositionDataSetId.
#  21-May-2013   jdw add deposit storage type
#  19-Sep-2013   jdw fix issues in ReferenceFileComponents and ReferenceFileInfo
#  26-Feb-2014   jdw adjust debugging output
#  10-Mar-2014   jdw add support for symbolic partition information.
#  28-Jun-2014   jdw make getPartitionNumberSearchTarget() public add  getVersionIdSearchTarget()
#   5-Jul-2014   jdw add getContentTypeSearchTarget()
#  14-Sep-2014   jdw add contentTypeExists(contentType)
#  29-Apr-2015   jdw add inline storage type for internal workflow variables -
#   7-May-2015   lm  add tempdep storage type
#  17-May-2015   jw  Restore method call with typo in name - getDepositonDataSetId()
#  30-Aug-2015   jw  Fix typo in method name in class ReferenceFileComponents
#   2-Sep-2015   jdw add accessor for supported storageTypes
#  23-Oct-2017   jdw add logging
##
"""
Classes defining references and naming conventions for data resources.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"
__version__ = "V0.01"

import os
import sys
import string
import glob
import traceback

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.config.ConfigInfoData import ConfigInfoData

import logging

logger = logging.getLogger(__name__)


class DataReferenceBase(object):

    """ Base class for data references such as files, web services and
        database services.

    """

    def __init__(self):
        self._referenceType = None

    def setReferenceType(self, refType):
        self._referenceType = refType

    def getReferenceType(self):
        return self._referenceType


class ReferenceFileInfo(object):

    """ Accessors for nomenclature conventions for reference files.
    """

    def __init__(self, verbose=False, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        #
        self.__ciD = ConfigInfoData(siteId=None, verbose=self.__verbose, log=self.__lfh).getConfigDictionary()
        #
        self.__contentD = self.__ciD["CONTENT_TYPE_DICTIONARY"]
        self.__formatD = self.__ciD["FILE_FORMAT_EXTENSION_DICTIONARY"]
        #
        self.__acronymD = {}
        for k, v in self.__contentD.items():
            self.__acronymD[v[1]] = k
        #
        self.__extD = {}
        for fmt, ext in self.__formatD.items():
            if ext not in self.__extD:
                self.__extD[ext] = []
            self.__extD[ext].append(fmt)
        #

    def contentTypeExists(self, contentType):
        try:
            return contentType in self.__contentD.keys()
        except Exception as _e:  # noqa: F841
            return False

    def getContentType(self, acronymName):
        try:
            return self.__acronymD[acronymName]
        except Exception as _e:  # noqa: F841
            return None

    def getFormatTypes(self, contentType):
        try:
            return self.__contentD[contentType][0]
        except Exception as _e:  # noqa: F841
            return []

    def getContentTypeAcronym(self, contentType):
        try:
            return self.__contentD[contentType][1]
        except Exception as _e:  # noqa: F841
            return None

    def getExtensionFormats(self, extension):
        try:
            return self.__extD[extension]
        except Exception as _e:  # noqa: F841
            return []

    def dump(self, ofh):
        ofh.write("\n+RefernceFileInfo.dump() content dictionary %r\n" % self.__contentD.items())

        for k in sorted(self.__acronymD.keys()):
            v = self.__acronymD[k]
            ofh.write("\n+RefernceFileInfo.dump() acronym %s contentType %s\n" % (k, v))

        ofh.write("\n+RefernceFileInfo.dump() extension dictionary %r\n" % self.__extD.items())


class ReferenceFileComponents(object):

    """ Provides methods for deconstructing  reference file names in terms of
        of their constituent attributes.
    """

    def __init__(self, fileName=None, verbose=False, log=sys.stderr):
        self.__fileName = fileName
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False
        #
        self.__reset()

        self.__rfI = ReferenceFileInfo(verbose=self.__verbose, log=self.__lfh)
        if self.__fileName is not None:
            self.__splitFileName()

    def __reset(self):
        self.__depositionDataSetId = None
        self.__filePartNumber = None
        self.__contentType = None
        self.__contentTypeAcronym = None
        self.__contentFormat = None
        self.__versionId = None

    def set(self, fileName):
        self.__fileName = fileName
        self.__reset()
        return self.__splitFileName()

    def get(self):
        """ Convenience method returning the elements of the reference file name
            in a single tuple.
        """
        return (self.__depositionDataSetId, self.__contentType, self.__contentFormat, self.__filePartNumber, self.__versionId)

    def __splitFileName(self):
        """ Internal method to decompose reference file name into components.
        """
        self.__reset()

        try:
            fFields = str(self.__fileName).strip().split(".")
            baseName = str(fFields[0]).strip()
            self.__formatExt = str(fFields[1]).strip()
            if len(fFields) > 2:
                self.__versionId = int(str(fFields[2][1:]))

            nFields = baseName.split("_")
            self.__depositionDataSetId = str(nFields[0] + "_" + nFields[1]).upper()
            #
            if self.__debug:
                logger.debug("+DataFileReferenceComponents.__splitFileName() fields %r", nFields)
            #
            if self.__rfI.contentTypeExists(str(nFields[2])):
                self.__contentTypeAcronym = str(nFields[2])
                self.__contentType = str(nFields[2])
            else:

                self.__contentTypeAcronym = str(nFields[2])
                self.__contentType = self.__rfI.getContentType(self.__contentTypeAcronym)
            #
            self.__filePartNumber = int(nFields[3][1:])

            fList = self.__rfI.getFormatTypes(self.__contentType)
            fmtList = self.__rfI.getExtensionFormats(self.__formatExt)
            for fmt in fmtList:
                if fmt in fList:
                    self.__contentFormat = fmt
            return True
        except Exception as e:
            if self.__debug:
                logger.debug("+DataFileReferenceComponents.__splitFileName() failed for %r %r", self.__fileName, str(e))
                traceback.print_exc(self.__lfh)
        return False

    def getVersionId(self):
        """ Return version identifier (integer), current symbolic setting, or None
        """
        return self.__versionId

    def getDepositionDataSetId(self):
        """ Return the data set identier -  (uppoer case)
        """
        return self.__depositionDataSetId

    def getDepositonDataSetId(self):
        """ Return the data set identier -  (uppoer case)
        """
        return self.__depositionDataSetId

    def getPartitionNumber(self):
        """  Return the file partition number (integer)  or symbolic setting.
        """
        return self.__filePartNumber

    def getContentTypeAcronym(self):
        """  Return the content type acronym
        """
        return self.__contentTypeAcronym

    def getContentType(self):
        """ Return the content type
        """
        return self.__contentType

    def getContentFormat(self):
        """ Return the content format using content acronym and file extension.
        """
        return self.__contentFormat


class DataFileReference(DataReferenceBase):

    """
    """

    def __init__(self, siteId=None, verbose=False, log=sys.stderr):
        super(DataFileReference, self).__init__()
        #
        self.__siteId = siteId
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        #
        self.__cI = ConfigInfo(siteId=self.__siteId, verbose=self.__verbose, log=self.__lfh)
        #
        self.__contentType = None
        """A supported content type:
           - model
           - structure-factors
           - nmr-restraints
           - em-volume
           - others to be enumerated
           """
        self.__fileFormat = None
        """A supported file format:
           - pdbx/mmcif
           - pdb
           - pdbml
           - nmr-star
           - others to be enumerated
           """
        self.__storageType = None
        """Storage type:
           - archive or wf-archive
           - wf-instance
           - wf-shared
           - deposit
           - tempdep
           - session or wf-session
           - inline
           - others to be enumerated
           """
        self.__versionId = None
        """Version identifier:
           - latest
           - orginal
           - next
           - previous
           - version number (1,2,...,)
           """
        #
        self.__filePartNumber = 1
        """Placeholder for future integer index for file partitioning.
        """
        #
        self.__depositionDataSetId = None
        """Deposition data set identifier (e.g. D_123456)
        """
        self.__workflowInstanceId = None
        """Workflow instance identifier (e.g. W_123456)
        """
        #
        self.__workflowNameSpace = None
        """Workflow name space identifier (alpha-numeric character string)
        """
        #
        self.__contentInfoD = self.__cI.get("CONTENT_TYPE_DICTIONARY")
        """Dictionary of supported file formats for each recognized content type.
           An acronym for each content type is included.
        """
        #
        self.__formatExtensionD = self.__cI.get("FILE_FORMAT_EXTENSION_DICTIONARY")
        """Dictionary of recognized file formats and file name extensions"""
        #
        self.__storageTypeList = ["archive", "autogroup", "wf-archive", "wf-instance", "wf-shared", "session", "wf-session", "deposit", "inline", "tempdep"]
        """List of supported storage types/locations"""
        #
        self.__depositionDataSetIdPrefix = "D_"
        """A depostion data set identifier begins with this prefix and is followed
           by a string of digits (e.g. D_123456789)"""
        #
        self.__groupDataSetIdPrefix = "G_"
        """A group data set identifier begins with this prefix and is followed
           by a string of digits (e.g. G_1234567)"""
        #
        self.__workflowInstanceIdPrefix = "W_"
        """A workflow instance identifier begins with this prefix and is followed
           by a string of digits (e.g. W_123456789)"""
        #
        self.__versionNameList = ["latest", "original", "previous", "next", "none"]
        self.__partitionNameList = ["latest", "original", "previous", "next", "none"]
        #
        #
        self.__externalFilePath = None
        """Placeholder for referencing a file name that is *external* to the archive
           or workflow system.  Setting this path implies a content type of *external*
           and other attributes of the reference will be treated as unknown/unassignable.
        """
        self.__sessionPath = "."
        """Optional path used as a file system directory for any files with 'session' storage type.

           The default value for the session storage is the current directory.
        """
        self.__sessionDataSetId = None
        """Optional session data set identifier (e.g. 1abc)
        """

        #

    def getSitePrefix(self):
        """Returns:

           Current setting of the site prefix.

        """
        return self.__cI.get("SITE_PREFIX")

    def setSessionPath(self, dirPath=None):
        """Set the full directory path for 'session' type storage.  The 'session' feature provides
           a means to support workflow file naming conventions for applications with transient
           storage requirements.

           Returns True for any non-null argument.  No check is performed for the existence of
           this path on input.
        """
        if dirPath is not None:
            self.__sessionPath = dirPath
            return True
        else:
            return False

    def setExternalFilePath(self, filePath, fileFormat="any"):
        """Set the full file path for this reference outside of the archive/workflow system.
           Other identifying attributes of this file reference are ignored/by-passed when
           this path is set.   This feature is provided to permit external data with alternative
           file name conventions to be used within data file references.

           File format may be optionall specified and must correspond to a supported
           format or the defaule 'any'.

           Returns:

           True if the assignment was successful or False otherwise.

        """
        if (filePath is None) or (len(filePath) < 1):
            return False

        if fileFormat in self.__formatExtensionD.keys():
            self.__fileFormat = fileFormat
        else:
            return False
        #
        # reset attributes -
        #
        self.__contentType = None
        self.__storageType = None
        self.__versionId = None
        self.__filePartNumber = 1
        self.__depositionDataSetId = None
        self.__workflowInstanceId = None
        self.__workflowNameSpace = None
        #
        self.__externalFilePath = None
        #
        try:
            self.__externalFilePath = os.path.abspath(filePath)
            (pth, fn) = os.path.split(self.__externalFilePath)
            if pth is None or fn is None:
                return False
            return True
        except Exception as _e:  # noqa: F841
            pass

        return False

    def setContentTypeAndFormat(self, contentType, fileFormat):
        """Set the content type and file format for the file reference.

           Examples of supported content types include:
           - model
           - structure-factors
           - nmr-restraints
           - nmr-chemical-shifts
           - component-image
           - component-definition
           - validation-report
           - em-volume

           Supported formats for each content type are defined in file format
           dictionary (`self.__contentInfoD`).

           Returns:

           True for a recognized content type  or False otherwise.
           """
        tS = str(contentType).lower()
        try:
            self.__contentInfoD.keys()
        except Exception as e:
            logger.exception("Failing with %r", str(e))

        if tS in self.__contentInfoD.keys():
            self.__contentType = tS
            fS = str(fileFormat).lower()
            if (fS in self.__contentInfoD[tS][0]) or ("any" in self.__contentInfoD[tS][0]):
                self.__contentType = tS
                self.__fileFormat = fS
                if self.__debug:
                    logger.debug("++setContentTypeAndFormat -- returning True with self.__contentType: %s", self.__contentType)
                    logger.debug("++setContentTypeAndFormat -- returning True with self.__fileFormat: %s", self.__fileFormat)
                self.setReferenceType("file")
                return True
            else:
                if self.__debug:
                    logger.debug("++setContentTypeAndFormat -- returning False with tS: %s", tS)
                    logger.debug("++setContentTypeAndFormat -- returning False with fS: %s", fS)
                return False
        else:
            if self.__debug:
                logger.debug("++setContentTypeAndFormat -- unrecognized cotentent type %r", tS)
            return False

    def getStorageTypeList(self):
        return self.__storageTypeList

    def setStorageType(self, storageType):
        """Set the storage type for this file reference.

           Supported storage types include:
           - archive or wf-archive
           - wf-instance
           - wf-shared
           - deposit
           - tempdep
           - session or wf-session

           Returns:

           True for a recognized storage type or False otherwise.

        """
        tS = str(storageType).lower()
        if tS in self.__storageTypeList:
            self.__storageType = tS
            if tS not in ["inline", "constant"]:
                self.setReferenceType("file")
            return True
        else:
            return False

    def setVersionId(self, versionId):
        """Set the version identifier for this file reference.

           Supported version identifiers include:
           - latest, ...
           - orginal
           - an integer version number (1,2,...,)

           Returns:

           True for a valid version identifier or False otherwise.

        """
        tS = str(versionId).lower()
        if versionId in self.__versionNameList:
            self.__versionId = tS
            return True
        elif self.__isInteger(tS):
            self.__versionId = tS
            return True
        else:
            return False

    def __isInteger(self, str_in):
        """ Is the given string an integer?	"""
        ok = True
        try:
            num = int(str_in)  # noqa: F841 pylint: disable=unused-variable
        except ValueError:
            ok = False
        return ok

    def setDepositionDataSetId(self, dId):
        """Set the deposition data set identifier.

        A depostion data set identifier begins with the prefix *D_* and is followed
        by a string of digits (e.g. D_123456789).

        Returns:

        True if the input identifier is a properly formed identifier
        or False otherwise.

        """
        tS = str(dId).upper()
        if (not tS.startswith(self.__depositionDataSetIdPrefix)) and (not self.__groupDataSetIdPrefix):
            return False
        tSL = tS.split("_")
        if (len(tSL) > 1) and self.__isInteger(tSL[1]):
            self.__depositionDataSetId = tS
            return True
        else:
            return False

    def setWorkflowInstanceId(self, wId):
        """Set the workflow instance identifier.

        A workflow instance identifier begins with the prefix *W_* and is followed
        by a string of digits (e.g. W_123456789)

        Returns:

        True if the input identifier is a properly formed identifier
        or False otherwise.

        """
        tS = str(wId).upper()
        if not tS.startswith(self.__workflowInstanceIdPrefix):
            return False
        tSL = tS.split("_")
        if (len(tSL) > 1) and self.__isInteger(tSL[1]):
            self.__workflowInstanceId = tS
            return True
        else:
            return False

    def setSessionDataSetId(self, sId):
        """Set the session data set identifier.

        Data set identifier applied for session storage. No conventions are
        assumed for this identifier.

        Returns:

        True if the input identifier is non-blank
        or False otherwise.

        """
        if sId is not None and len(sId) > 0:
            self.__sessionDataSetId = str(sId).upper()
            return True
        else:
            return False

    def setWorkflowNameSpace(self, wNameSpace):
        """Set the workflow name space identifier.

           This identifier must be an alpha numeric string containing only
           characters [a-zA-Z0-9].

           Returns:

           True if the input identifier is a properly formed identifier
           or False otherwise.

        """
        if (wNameSpace is None) or (len(str(wNameSpace)) < 1):
            return False
        for cv in str(wNameSpace):
            if (cv not in string.ascii_letters) and (cv not in string.digits):
                return False
        self.__workflowNameSpace = wNameSpace
        return True

    def setPartitionNumber(self, iPartitionNumber=1):
        """Set the integer file partition number.  This is used to identify the physical
           pieces of a single logical data file.

           Supported values for partition include:
           - latest, ...
           - orginal
           - an integer version number (1,2,...,)

           Returns:

           True if the input partition is properly formed or False otherwise.

        """
        ok = False
        try:
            tS = str(iPartitionNumber).lower()
            if iPartitionNumber in self.__partitionNameList:
                self.__filePartNumber = tS
                ok = True
            elif self.__isInteger(tS):
                self.__filePartNumber = int(tS)
                ok = True
            else:
                ok = False
        except Exception:
            ok = False
        if self.__debug:
            logger.debug("+DataFileReference.setPartitionNumber() setting is  %r", self.__filePartNumber)
        return ok

    def getPartitionNumber(self):
        """Returns:

           The current partition number  or *1* if this is not set.
        """
        return self.__filePartNumber

    def getContentType(self):
        """Returns:

           The current content type or *None* if this is not set.
        """
        return self.__contentType

    def getFileFormat(self):
        """Returns:

           The current file format or *None* if this is not set.
        """
        return self.__fileFormat

    def getStorageType(self):
        """Returns:

           The current storage type or *None* if this is not set.
        """

        return self.__storageType

    def getVersionId(self):
        """Returns:

           The current version identifier or *None* if this is not set.
        """
        return self.__versionId

    def getDepositionDataSetId(self):
        """Returns:

           The current deposition data set identifier  or *None* if this is not set.
        """
        return self.__depositionDataSetId

    def getWorkflowInstanceId(self):
        """Returns:

           The current workflow instance identifier  or *None* if this is not set.
        """
        return self.__workflowInstanceId

    def getWorkflowNameSpace(self):
        """Returns:

           The current workflow name space identifier  or *None* if this is not set.
        """
        return self.__workflowNameSpace

    #
    # ------------------------------------------------------------------------------------------------------------------------------------
    #
    # --- The following public methods derive information from the settings in the previous methods --
    #

    def isReferenceValid(self):
        """Test if the reference information is complete and the data reference is valid.

           Valid references are:

           - A path external to the archive/worflow system
           - A fully defined internal reference consisting or identifiers,
             content type, storage type, format, and version.

           Note that this is NOT an existence test.  References may be defined and validated
           before the file objects which they reference are created.

           Returns:

           True for either a valid external or internal reference or False otherwise.

        """
        if self.__externalFilePath is not None:
            return True
        else:
            return self.__isInternalReferenceValid()

    def getDirPathReference(self):
        """ Get the path to the directory containing the data file reference.

            Returns:

            The file system path to the directory containing the file reference or *None*
            if this cannot be determined.

        """
        if self.__externalFilePath is not None:
            return self.__externalFilePath
        # if (self.__isInternalReferenceValid()):
        #    return self.__getInternalPath()
        else:
            return self.__getInternalPath()

    def getFilePathReference(self):
        """Get the versioned file path for an internal data file reference or the path
           to an external data file reference.

           Returns:

           The file system path to the file reference or *None* if this cannot be determined.

        """
        if self.__externalFilePath is not None:
            return self.__externalFilePath
        if not self.__isInternalReferenceValid():
            return None
        return self.__getInternalFilePath()

    def getFilePathExists(self, fP):
        try:
            if os.access(fP, os.R_OK):
                return True
            else:
                return False
        except Exception as _e:  # noqa: F841
            if self.__verbose:
                traceback.print_exc(self.__lfh)
            return False

    def getFileVersionNumber(self):
        """Get the version number corresponding to the current data file reference.

           Returns:

           The version number 1-N of the current data reference or 0 otherwise.
           External references are treated as having no version and 0 is returned for
           these cases.

        """

        if self.__externalFilePath is not None:
            return 0
        if not self.__isInternalReferenceValid():
            return 0
        return self.__getInternalVersionNumber()

    #
    # ------------------------------------------------------------------------------------------------------------------------------------
    #
    # --- The following private worker methods support the public path and validation methods.
    #

    def __isInternalReferenceValid(self):
        """Test if the current reference information is complete for an internal reference.
           A reference is considered internal which points within the archive, workflow
           instance, deposit or session file systems.  Otherwise the reference is considered external
           and not subject to internal naming or path conventions.

           Note that this is NOT an existence test.  References may be defined and validated
           before the file objects which they reference are created.

           Returns:

           True if the internal reference is complete or False otherwise.
        """
        referenceType = self.getReferenceType()

        if referenceType == "file":
            if (self.__contentType is None) or (self.__fileFormat is None) or (self.__storageType is None) or (self.__versionId is None):
                # logger.debug("self.__contentType is: %s", self.__contentType)
                # logger.debug("self.__fileFormat is: %s", self.__fileFormat)
                # logger.debug("self.__storageType is: %s", self.__storageType)
                # logger.debug("self.__versionId is: %s", self.__versionId)

                return False

            if (self.__storageType in ["archive", "autogroup", "wf-archive", "wf-instance", "wf-shared", "deposit", "tempdep"]) and (self.__depositionDataSetId is None):
                logger.debug("self.__depositionDataSetId is: %s", self.__depositionDataSetId)
                return False

            if (self.__storageType in ["session", "wf-session"]) and (self.__sessionDataSetId is None):
                return False

            if (self.__storageType == "wf-instance") and (self.__workflowInstanceId is None):
                return False

            if (self.__storageType == "wf-shared") and (self.__workflowNameSpace is None):
                return False

            return True

        elif referenceType == "directory":
            if self.__storageType is None:
                return False

            if (self.__storageType in ["archive", "autogroup", "wf-archive", "wf-instance", "wf-shared", "deposit", "tempdep"]) and (self.__depositionDataSetId is None):
                return False

            if (self.__storageType == "wf-instance") and (self.__workflowInstanceId is None):
                return False

            if (self.__storageType == "wf-shared") and (self.__workflowNameSpace is None):
                return False

            if (self.__storageType in ["wf-session", "session"]) and (self.__sessionPath is None):
                return False

            return True

        else:
            return False

    def __getExternalPath(self):
        """Get the path of the current external file reference.

           Returns:

           The external file path.  *None* is returned on failure.
        """
        try:
            (pth, _fn) = os.path.split(self.__externalFilePath)
            return pth
        except Exception as _e:  # noqa: F841
            return None

    def __getExternalFileNameBase(self):
        """Get the base file name for the current external file reference.

           Returns:

           The external base file name.  *None* is returned on failure.
        """
        try:
            (_pth, fn) = os.path.split(self.__externalFilePath)
            return fn
        except Exception as _e:  # noqa: F841
            return None

    def __getInternalPath(self):
        """Compute the path to the current file reference within the archive/workflow file system.

        The file path convention is:
        - archive files     = <SITE_ARCHIVE_STORAGE_PATH>/archive/<deposition data set id>/
        - deposit files     = <SITE_DEPOSIT_STORAGE_PATH>/archive/<deposition data set id>/
        - temp deposit files     = <SITE_TEMPDEP_STORAGE_PATH>/tempdep/<deposition data set id>/
        - workflow shared   = <SITE_WORKFLOW_STORAGE_PATH>/workflow/<deposition data set id>/shared/<self.__workflowNameSpace>
        - workflow instance = <SITE_WORKFLOW_STORAGE_PATH>/workflow/<deposition data set id>/instance/<self.__workflowInstanceId>
        - session files     = session path/

        Top-level site-specific path details are obtained from the SiteInfo() class.

        Returns:

        The path of the directory containing this data file reference.  *None* is returned on failure.

        """
        try:
            if self.__storageType == "archive" or self.__storageType == "wf-archive":
                tpth = os.path.join(self.__cI.get("SITE_ARCHIVE_STORAGE_PATH"), "archive", self.__depositionDataSetId)
            elif self.__storageType == "autogroup":
                tpth = os.path.join(self.__cI.get("SITE_DEPOSIT_STORAGE_PATH"), "autogroup", self.__depositionDataSetId)
            elif self.__storageType == "deposit":
                tpth = os.path.join(self.__cI.get("SITE_DEPOSIT_STORAGE_PATH"), "deposit", self.__depositionDataSetId)
            elif self.__storageType == "tempdep":
                tpth = os.path.join(self.__cI.get("SITE_TEMPDEP_STORAGE_PATH"), "tempdep", self.__depositionDataSetId)
            elif self.__storageType == "wf-shared":
                tpth = os.path.join(self.__cI.get("SITE_WORKFLOW_STORAGE_PATH"), "workflow", self.__depositionDataSetId, "shared", self.__workflowNameSpace)
            elif self.__storageType == "wf-instance":
                tpth = os.path.join(self.__cI.get("SITE_WORKFLOW_STORAGE_PATH"), "workflow", self.__depositionDataSetId, "instance", self.__workflowInstanceId)
            elif self.__storageType in ["session", "wf-session"]:
                tpth = self.__sessionPath
            else:
                tpth = None
            pth = os.path.abspath(tpth)
        except Exception as e:
            logger.exception("Failing with %r", str(e))

            pth = None

        return pth

    def __getInternalFileNameBase(self):
        """Compute the base file name based on the current values of storage type, identifer, content type, file format.

        The file name convention is:
        - archive/shared  files  = <deposition data set id>_<content acronym>_<part number>.<format_extenstion>
        - instance files         = <deposition data set id>_<content acronym>_<part number>.<format_extension>
        - session  files         = <session data set id>_<content acronym>_<part number>.<format_extenstion>

        Returns:

        The base file name. This base file name lacks version details.

        """

        try:
            if self.getReferenceType() != "file":
                return None

            if self.__storageType in ["archive", "autogroup", "wf-archive", "wf-shared", "deposit", "tempdep"]:
                fn = (
                    self.__depositionDataSetId
                    + "_"
                    + self.__contentInfoD[self.__contentType][1]
                    + "_P"
                    + str(self.__filePartNumber)
                    + "."
                    + self.__formatExtensionD[self.__fileFormat]
                )
            elif self.__storageType in ["session", "wf-session"]:
                fn = (
                    self.__sessionDataSetId
                    + "_"
                    + self.__contentInfoD[self.__contentType][1]
                    + "_P"
                    + str(self.__filePartNumber)
                    + "."
                    + self.__formatExtensionD[self.__fileFormat]
                )
            elif self.__storageType in ["wf-instance"]:
                fn = (
                    self.__depositionDataSetId
                    + "_"
                    + self.__contentInfoD[self.__contentType][1]
                    + "_P"
                    + str(self.__filePartNumber)
                    + "."
                    + self.__formatExtensionD[self.__fileFormat]
                )
            else:
                fn = None
        except Exception as e:
            logger.exception("Failing with %r", str(e))
            fn = None

        return fn

    def __getInternalFilePath(self):
        """Compute the versioned file path for a file within the archive/worflow file system.

           If either the *latest*, *next*, or *previous* version of the referenced file is
           selected then a file system check is performed to determine the appropriate
           version number.

        Returns:

        File path including version or None on failure.
        """

        try:
            if self.getReferenceType() != "file":
                return None
            dirPath = self.__getInternalPath()
            fN = self.__getInternalFileNameVersioned()
            pth = os.path.join(dirPath, fN)
            return pth
        except Exception as _e:  # noqa: F841
            return None

    def getVersionIdSearchTarget(self):
        """  Create a search target for the files containing any version identifier consistent with the
             current file settings.

             Returns a the search target appropriate for glob() or None
        """
        try:
            if self.getReferenceType() != "file":
                return None
            baseName = self.__getInternalFileNameBase()
            vst = baseName + ".V*"
            return vst
        except Exception as _e:  # noqa: F841
            return None

    def __getInternalFileNameVersioned(self):
        """Compute the versioned file name for a file within the archive/worflow file system.

           If either the *latest*, *next*, or *previous* version of the referenced file is
           selected then a file system check is performed to determine the appropriate
           version number.

        Returns:

        File name including version or None on failure.
        """
        try:
            if self.getReferenceType() != "file":
                return None
            dirPath = self.__getInternalPath()
            #
            # First resolve any symbolic partition information -
            #
            self.__filePartNumber = self.__getInternalPartitionNumber()
            #
            baseName = self.__getInternalFileNameBase()
            if self.__versionId == "latest":
                iV = self.__latestVersion(dirPath, baseName)
                if iV == 0:
                    # No version exists so start at 1
                    fn = baseName + ".V1"
                else:
                    #
                    fn = baseName + ".V" + str(int(iV))

            elif self.__versionId == "next":
                iV = self.__latestVersion(dirPath, baseName)
                if iV == 0:
                    # No version exists so start at 1
                    fn = baseName + ".V1"
                else:
                    #
                    fn = baseName + ".V" + str(int(iV + 1))

            elif self.__versionId == "previous":
                iV = self.__latestVersion(dirPath, baseName)
                if iV <= 1:
                    # No previous version.
                    fn = None
                else:
                    #
                    fn = baseName + ".V" + str(int(iV - 1))

            elif self.__versionId == "original":
                fn = baseName + ".V1"
            elif self.__versionId == "none":
                fn = baseName
            else:
                fn = baseName + ".V" + str(int(self.__versionId))

            return fn
        except Exception as e:
            logger.exception("failure in getInternalFileNameVersioned %r", str(e))
            return None

    def __getInternalVersionNumber(self):
        """Determine the version number corresponding to the current version Id setting.

           If either the *latest*, *next*, or *previous* version of the referenced file is
           selected then a file system check is performed to determine the appropriate
           version number.

        Returns:

        Return a version number from 1-N   or 0 failure.
        """
        try:
            if self.getReferenceType() != "file":
                return 0
            dirPath = self.__getInternalPath()
            self.__filePartNumber = self.__getInternalPartitionNumber()
            baseName = self.__getInternalFileNameBase()
            if self.__versionId == "latest":
                iV = self.__latestVersion(dirPath, baseName)
            elif self.__versionId == "next":
                iV = self.__latestVersion(dirPath, baseName)
                iV += 1
            elif self.__versionId == "previous":
                iV = self.__latestVersion(dirPath, baseName)
                iV -= 1
                if iV < 0:
                    iV = 0
            elif self.__versionId == "original":
                iV = 1
            else:
                iV = int(self.__versionId)
            return iV
        except Exception as e:
            if self.__debug:
                logger.exception("Failing with %r", str(e))
        return 0

    def __latestVersion(self, dirPath, baseName):
        """ Get the latest version of file *baseName* in path *dirPath*.

            The convention for version numbering is <baseName>.V#

            Returns:

            The latest integer version number  or 0 if no versions exist.

        """
        try:
            fN = None
            if self.getReferenceType() != "file":
                return 0
            vList = []
            fileList = os.listdir(dirPath)
            for fN in fileList:
                # logger.debug("__latestVersion - baseName %s fN %s", baseName,fN)
                if fN.startswith(baseName):
                    fSp = fN.split(".V")
                    if (len(fSp) < 2) or (not fSp[1].isdigit()):
                        continue
                    vList.append(int(fSp[1]))
                    # logger.debug("__latestVersion - vList %r\n" % (fSp))
            if len(vList) > 0:
                vList.sort()
                return vList[-1]
            else:
                return 0
        except Exception as e:
            if self.__debug:
                logger.exception("Failing -dirPath %s  baseName %s fN %s with %s", dirPath, baseName, fN, str(e))

        return 0

    ##
    ##
    def __latestPartitionNumber(self, dirPath, searchTarget):
        """ Get the latest partition number of file in path *dirPath*
            consistent with current file settings.

            Returns:

            The latest integer partition number  or 0 if no files exist.

        """
        try:
            fN = None
            if self.getReferenceType() != "file":
                return 0
            pList = []
            searchPath = os.path.join(dirPath, searchTarget)
            if self.__debug:
                logger.debug("+DataFileReference.__lastestPartitionNumber() search target %s", searchPath)
            pathList = glob.glob(searchPath)
            for pth in pathList:
                if self.__debug:
                    logger.debug("+DataFileReference.__lastestPartitionNumber() search path %s", pth)
                (_td, fN) = os.path.split(pth)
                fL1 = fN.split(".")
                fL2 = fL1[0].split("_")
                pList.append(int(fL2[3][1:]))

            if self.__debug:
                logger.debug("+DataFileReference.__lastestPartitionNumber() part number list  %r", pList)
            if len(pList) > 0:
                pList.sort()
                return pList[-1]
            else:
                return 0
        except Exception as e:
            if self.__debug:
                logger.exception("Failing with %r", str(e))

        return 0

    def getPartitionNumberSearchTarget(self):
        """Create a search target for the files containing any partition number consistent with the
           current file settings.   The seach target is independent of version identifier.

        The file name convention is:
        - archive/shared  files  = <deposition data set id>_<content acronym>_P*.<format_extenstion>
        - instance files         = <deposition data set id>_<content acronym>_P*.<format_extension>
        - session  files         = <session data set id>_<content acronym>_P*.<format_extenstion>

        Returns:

        A search string appopriate for glob().

        """

        try:
            if self.getReferenceType() != "file":
                return None

            if self.__storageType in ["archive", "autogroup", "wf-archive", "wf-shared", "deposit", "tempdep"]:
                fn = self.__depositionDataSetId + "_" + self.__contentInfoD[self.__contentType][1] + "_P*" + "." + self.__formatExtensionD[self.__fileFormat] + "*"
            elif self.__storageType in ["session", "wf-session"]:
                fn = self.__sessionDataSetId + "_" + self.__contentInfoD[self.__contentType][1] + "_P*" + "." + self.__formatExtensionD[self.__fileFormat] + "*"
            elif self.__storageType in ["wf-instance"]:
                fn = self.__depositionDataSetId + "_" + self.__contentInfoD[self.__contentType][1] + "_P*" + "." + self.__formatExtensionD[self.__fileFormat] + "*"
            else:
                fn = None
        except Exception as e:
            logger.exception("Failing with %r", str(e))

            fn = None

        return fn

    def getContentTypeSearchTarget(self):
        """Create a search target for the files containing any variation consistent with the
           content type in current file settings.   The seach target is independent of partition,
           format and version identifier.

        The file name convention is:
        - archive/shared  files  = <deposition data set id>_<content acronym>_P*
        - instance files         = <deposition data set id>_<content acronym>_P*
        - session  files         = <session data set id>_<content acronym>_P*

        Returns:

        A search string appopriate for glob().

        """

        try:
            # if (self.getReferenceType() != 'file'):
            #  return None

            if self.__storageType in ["archive", "autogroup", "wf-archive", "wf-shared", "deposit", "tempdep"]:
                fn = self.__depositionDataSetId + "_" + self.__contentInfoD[self.__contentType][1] + "_P*"
            elif self.__storageType in ["session", "wf-session"]:
                fn = self.__sessionDataSetId + "_" + self.__contentInfoD[self.__contentType][1] + "_P*"
            elif self.__storageType in ["wf-instance"]:
                fn = self.__depositionDataSetId + "_" + self.__contentInfoD[self.__contentType][1] + "_P*"
            else:
                fn = None
        except Exception as e:
            logger.exception("Failing storage %r data set id %r content type %r with %r", self.__storageType, self.__depositionDataSetId, self.__contentType, str(e))
            fn = None

        return fn

    def __getInternalPartitionNumber(self):
        """Determine the partition number corresponding to the current partition number setting.

           If either the *latest*, *next*, or *previous* version of the referenced file is
           selected then a file system check is performed to determine the appropriate
           partition number.

        Returns:

        Return a partition number from 1-N   or 0 failure.
        """
        try:
            if self.getReferenceType() != "file":
                return 0
            dirPath = self.__getInternalPath()
            searchTarget = self.getPartitionNumberSearchTarget()
            if self.__filePartNumber == "latest":
                iP = self.__latestPartitionNumber(dirPath, searchTarget)
            elif self.__filePartNumber == "next":
                iP = self.__latestPartitionNumber(dirPath, searchTarget)
                iP += 1
            elif self.__filePartNumber == "previous":
                iP = self.__latestPartitionNumber(dirPath, searchTarget)
                iP -= 1
                if iP < 0:
                    iP = 0
            elif self.__filePartNumber == "original":
                iP = 1
            else:
                iP = int(self.__filePartNumber)
            return iP
        except Exception as e:
            logger.exception("Failing with %r", str(e))

        return 0



import os
import os.path
import logging

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.io.locator.ReleaseFileNames import ReleaseFileNames

logger = logging.getLogger(__name__)


class LocalFTPPathInfo(object):
    def __init__(self, siteId=None):
        self.__siteId = siteId
        self.__cI = ConfigInfo(siteId=self.__siteId)

        self.ftp_pdb_root = self.__cI.get('SITE_PDB_FTP_ROOT_DIR')
        self.ftp_pdb = os.path.join(self.ftp_pdb_root, 'pdb', 'data', 'structures', 'all')
        self.ftp_emdb_root = self.__cI.get('SITE_EMDB_FTP_ROOT_DIR')
        self.ftp_emdb = os.path.join(self.ftp_emdb_root, 'emdb', 'structures')
        self.__mapping = {
            'model': 'mmCIF',
            'structure_factors': 'structure_factors',
            'chemical_shifts': 'nmr_chemical_shifts',
            'nmr_data': 'nmr_data',
        }

    def __get_mapping(self, file_type):
        return self.__mapping.get(file_type)

    def get_ftp_pdb(self):
        return self.ftp_pdb

    def get_ftp_emdb(self):
        return self.ftp_emdb

    def get_model_path(self):
        return os.path.join(self.ftp_pdb, self.__get_mapping('model'))

    def get_sf_path(self):
        return os.path.join(self.ftp_pdb, self.__get_mapping('structure_factors'))

    def get_cs_path(self):
        return os.path.join(self.ftp_pdb, self.__get_mapping('chemical_shifts'))

    def get_nmr_data_path(self):
        return os.path.join(self.ftp_pdb, self.__get_mapping('nmr_data'))

    def get_model_fname(self, accession):
        model_file_name = ReleaseFileNames().get_model(accession=accession, for_release=False)
        return os.path.join(self.get_model_path(), model_file_name)

    def get_structure_factors_fname(self, accession):
        sf_file_name = ReleaseFileNames().get_structure_factor(accession=accession, for_release=False)
        return os.path.join(self.get_sf_path(), sf_file_name)

    def get_chemical_shifts_fname(self, accession):
        cs_file_name = ReleaseFileNames().get_chemical_shifts(accession=accession, for_release=False)
        return os.path.join(self.get_cs_path(), cs_file_name)

    def get_nmr_data_fname(self, accession):
        nmr_data_file_name = ReleaseFileNames().get_nmr_data(accession=accession, for_release=False)
        return os.path.join(self.get_nmr_data_path(), nmr_data_file_name)



##
# File:  ReleasePathInfo.py
# Date:  11-Nov-2019  E. Peisach
#
# Updated:
##
"""
Common methods for finding path information for release directories used in the system.

"""
__docformat__ = "restructuredtext en"
__author__ = "Ezra Peisach"
__email__ = "peisach@rcsb.rutgers.edu"
__license__ = "Apache 2.0"
__version__ = "V0.07"

import os
import os.path
import logging

from wwpdb.utils.config.ConfigInfo import ConfigInfo

logger = logging.getLogger(__name__)


class ReleasePathInfo(object):
    def __init__(self, siteId=None):
        self.__siteId = siteId
        self.__cI = ConfigInfo(siteId=self.__siteId)

    def getForReleasePath(self, subdir=None, version="current", accession=None, em_sub_path=None):
        """Returns path to for-release directory.

        Input Parameters:
        subdir: If specified,  must be one of "added", "modified", "obsolete",
                "emd", "val-reports".
        version: "current" or "previous"

        returns path, or raises exception
        """
        basedir = os.path.join(self.__cI.get("SITE_ARCHIVE_STORAGE_PATH"), "for_release")

        if version not in ["current", "previous"]:
            raise NameError("version %s not allowed" % version)

        if version == "previous":
            basedir = os.path.join(basedir, "previous")

        if subdir:
            if subdir not in [
                "added",
                "modified",
                "obsolete",
                "emd",
                "val_reports",
                "em_val_reports",
            ]:
                raise NameError("subdir %s not allowed" % subdir)

            basedir = os.path.join(basedir, subdir)

            if em_sub_path and accession:
                if em_sub_path not in [
                    "header",
                    "map",
                    "fsc",
                    "images",
                    "masks",
                    "other",
                    "validation",
                ]:
                    raise NameError("em_sub_path %s not allowed" % em_sub_path)

                basedir = os.path.join(basedir, accession.upper(), em_sub_path)

        return basedir
